from __future__ import annotations
from pathlib import PurePath
from typing import List, Dict, TYPE_CHECKING


import time
import datetime
import logging
import traceback
from buffy.buffyserver.backend.storage.interface import StorageInterface
from buffy.buffyserver.api.v1.models import (
    Request,
    Response,
    ResponseDownloadStats,
)
from buffy.buffyserver.api.v1.models_recaching_strategies import (
    ReCachingStrategy,
    BaseReCachingStrategy,
)
from buffy.buffyserver.backend.request_download_service.queue import (
    RequestDownloadQueue,
    Download,
)

from buffy.buffyserver.backend.storage_janitor import StorageJanitor


log = logging.getLogger(__name__)


class RequestDownloaderService:
    server_tick_min_duration_sec: float = 0.3
    clean_storage_every_n_sec: int = 60
    validate_storage_at_boot: bool = True
    validate_storage_every_n_sec: int = 43200
    # 43200s = 12h
    def __init__(
        self,
        storage: StorageInterface,
        max_simultaneous_downloads: int = 4,
        max_simultaneous_downloads_per_domain: int = 1,
    ):
        self.storage: StorageInterface = storage
        self.downloads: RequestDownloadQueue = RequestDownloadQueue()
        self.max_simultaneous_downloads: int = max_simultaneous_downloads
        self.max_simultaneous_downloads_per_domain: int = (
            max_simultaneous_downloads_per_domain
        )
        self._strategies_states: Dict[str, Dict] = {}
        self.running: bool = False

        self._init_strategy_states()
        if not self.storage.test_connection():
            raise SystemError("Storage backend connection not successful")

        self._janitor = StorageJanitor(storage=self.storage)
        self._last_storage_clean: float = 0.0
        self._last_storage_validation: float = None

    def run(self):
        self._last_storage_validation: float = (
            0.0 if self.validate_storage_at_boot else time.time()
        )
        self.running = True
        while self.running:
            self._clean_and_validate_storage()
            for request in reversed(self.storage.list_requests()):
                if self._local_response_cache_needs_recaching(request):
                    self._create_new_response(request)
                if request.latest_requests:
                    # we reset the list request timings. client can check now if their request was allready processed by the backend. They just need to check if their timestamp is still `latest_requests` aka the wait line
                    request.latest_requests = []
                    request = self.storage.update_request(request)
            self._start_next_response_download()
            self._update_running_response_download_statuses()
            self._dispose_finished_response_downloads()

            time.sleep(self.server_tick_min_duration_sec)

    def _clean_and_validate_storage(self):
        current_time = time.time()
        if current_time - self._last_storage_clean > self.clean_storage_every_n_sec:
            log.debug("Clean storage...")
            self._janitor.cleanup_storage()
            log.debug("...cleaning storage done.")
            self._last_storage_clean = time.time()
        if (
            current_time - self._last_storage_validation
            > self.validate_storage_every_n_sec
        ):
            log.info("Validate storage...")
            self._janitor.verifiy_storage()
            log.info("...validating storage done.")
            self._last_storage_validation = time.time()

    def _init_strategy_states(self):
        for strategy in ReCachingStrategy.list():
            self._strategies_states[strategy.__name__] = {}

    def _local_response_cache_needs_recaching(self, request: Request) -> bool:
        """Depeding on the configured 'ReCachingStrategy' for a request, we determine if the request needs an updated response (aka "does the remote file needs to be downloaded (again)?")

        Args:
            request (Request): _description_

        Returns:
            bool: _description_
        """

        latest_response: Response = self.storage.get_response(
            request_id=request.id, version="latest"
        )

        if not latest_response:
            return True
        if TYPE_CHECKING:
            request.cache_configuration.recaching_strategy: BaseReCachingStrategy = (
                request.cache_configuration.recaching_strategy
            )
        strategy_name = request.cache_configuration.recaching_strategy.strategy_name
        return request.cache_configuration.recaching_strategy.request_needs_recache(
            request=request,
            latest_response=latest_response,
            strategy_state=self._strategies_states[strategy_name],
            storage=self.storage,
        )

    def _create_new_response(self, request: Request):
        version = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        previous_respond: Response = None

        # Register next version in version chain
        previous_respond = self.storage.get_response(
            request_id=request.id, version="latest"
        )
        if previous_respond:
            previous_respond.next_version = version
            self.storage.update_response(response=previous_respond)

        response = self.storage.create_response(
            response=Response(
                request_id=request.id,
                status="wait",
                version=version,
                previous_version=previous_respond.version if previous_respond else None,
                status_bytes_downloaded=0,
                request_datetime_utc=request.latest_request_datetime_utc,
                # content_attributes=downloader.get_response_content_attrs(),
            )
        )

        self.downloads.add_download(
            request=request,
            response=response,
            file_target=self.storage.write_response_content(response=response),
        )

    def _start_next_response_download(self):
        for domain in self.downloads.domains:
            if (
                len(self.downloads.get_downloads(by_status="run"))
                >= self.max_simultaneous_downloads
            ):
                return
            elif (
                len(self.downloads.get_downloads(by_domain=domain, by_status="run"))
                >= self.max_simultaneous_downloads_per_domain
            ):
                continue
            elif (
                len(self.downloads.get_downloads(by_domain=domain, by_status="wait"))
                == 0
            ):
                continue
            else:
                waiting_downloads: List[Download] = self.downloads.get_downloads(
                    by_domain=domain, by_status="wait"
                )
                waiting_downloads[0].thread.start()
                log.debug(f"STARTED DOWNLOAD {waiting_downloads[0].request}")

    def _update_running_response_download_statuses(self):
        running_downloads: List[Download] = self.downloads.get_downloads(
            by_status="run"
        )
        for download in running_downloads:
            download.response.download_stats = download.downloader.status
            download.response.status = "in_progress"
            if (
                download.response.content_attributes is None
                or download.response.content_attributes.is_empty()
            ):
                download.response.content_attributes = (
                    download.downloader.get_response_content_attrs(cached_only=True)
                )
            self.storage.update_response(download.response)

    def _dispose_finished_response_downloads(self):
        finished_downloads: List[Download] = self.downloads.get_downloads(
            by_status="finished"
        )
        download_to_clean: List[Download] = []
        for download in finished_downloads:
            tb = None
            try:
                download.thread.join(timeout=0.6)
            except Exception as e:
                tb = traceback.format_exc()
                log.error(tb)
                download.thread.failed = True
            if download.thread.failed:
                download.response.status = "failed"
                if not download.response.download_stats:
                    download.response.download_stats = ResponseDownloadStats()
                download.response.download_stats.error = (
                    tb if tb else download.thread.error_stack_trace
                )
            else:
                download.response.status = "ready"
                download.response.cached_datetime_utc = datetime.datetime.utcnow()
                download.response.download_stats = download.downloader.status
                download.response.content_hash_hex = download.downloader.hash
                download.response.content_download_path = str(
                    PurePath(
                        "request",
                        download.request.id,
                        "response",
                        "v",
                        download.response.version,
                        "content",
                    )
                )
                download.response.content_attributes = (
                    download.downloader.get_response_content_attrs(cached_only=True)
                )

            self.storage.update_response(download.response)
            download_to_clean.append(download)
        for download in download_to_clean:
            self.downloads.remove_download(download=download)



import os
import sys

SCRIPT_DIR = "."
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))


from buffy.buffypyclient import BuffyPyClient

c = BuffyPyClient()
req = c.create_request(url="https://www.foaas.com/awesome/:tim")
cached_responses = req.list_cached_versions()
for resp in cached_responses:
    print(resp.id)



import os
from subprocess import TimeoutExpired
import sys
import time
from shittywebserver import run_shitty_webserver
import json
import atexit
import shutil
from pathlib import Path
import multiprocessing
from typing import List

print("# Create server instances...")
SCRIPT_DIR = "."
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))
STORAGE_BASE_DIR = "./tests/tmp"
STORAGE_CLIENT = f"{STORAGE_BASE_DIR}/client"
STORAGE_SERVER = f"{STORAGE_BASE_DIR}/server"
CLEAN_RESULT_FILES_AT_EXIT: bool = True
BUFFY_SERVER_URL = "http://localhost:8008"
os.makedirs(STORAGE_CLIENT, exist_ok=True)
os.makedirs(STORAGE_SERVER, exist_ok=True)


SHITTY_WEBSERVER_BASE_URL = "http://localhost:8088/v1"

REDIS_HOST = os.getenv("CI_REDIS_HOST", "localhost")
print("REDIS_HOST", REDIS_HOST)
STORAGE_CONF = {"host": REDIS_HOST}
LOG_LEVEL = "DEBUG"


def end(exit=False):
    webserver_proc.kill()
    api_proc.kill()
    backend_proc.kill()
    if CLEAN_RESULT_FILES_AT_EXIT and os.path.isdir(STORAGE_BASE_DIR):
        print(f"CLEAR '{STORAGE_BASE_DIR}'")
        shutil.rmtree(STORAGE_BASE_DIR)
    if exit:
        exit()


atexit.register(end)

from buffy.buffyserver.api.v1.models import Request_in, Request, Response
from buffy.buffyserver.main import start_buffy_server
from buffy.buffypyclient import (
    BuffyPyClient,
    RequestCacheConfiguration,
    ReCachingStrategy,
)
from buffy.tools.utils import hashfile


print("Start shittywebserber as test endpoint...")
webserver_proc = run_shitty_webserver(port=8088, run_in_subprocess=True)
time.sleep(1)
print("Start Buffy instance...")

backend_proc, api_proc = start_buffy_server(
    env_vars={
        "CONFIGS_STORAGE_BACKEND_CONFIG": json.dumps(STORAGE_CONF),
        "CONFIGS_STORAGE_BACKEND_FILE_DIR": STORAGE_SERVER,
        "CONFIGS_LOG_LEVEL": LOG_LEVEL,
    },
    watchdog=False,
)
time.sleep(1)


def run_test_func(func, timeout_sec=60):
    print(f"## RUN TEST '{func.__name__}' with timeout of {timeout_sec} secs ")

    def check_backend_health():
        if not backend_proc.is_alive():
            backend_proc.join()
            return Exception(f"Backend error exitcode: {backend_proc.exitcode}")
        if not api_proc.is_alive():
            api_proc.join()
            return Exception(f"API Server error exitcode: {api_proc.exitcode}")

    start_time = time.time()
    p = multiprocessing.Process(target=func)
    p.start()
    while p.is_alive():
        server_exc = check_backend_health()
        if server_exc:
            p.terminate()
            p.join()
            raise server_exc
        if time.time() - start_time > timeout_sec:
            p.terminate()
            p.join()
            raise TimeoutError()
        time.sleep(0.1)
    p.join()
    if p.exitcode > 0:
        raise ValueError(f"Process exited with code {p.exitcode}")

    return


try:
    print("# Run tests...")
    bclient = BuffyPyClient(BUFFY_SERVER_URL)

    def test_client_params():
        c = BuffyPyClient(host="localhost", ssl=False)
        assert c.api_url == f"{BUFFY_SERVER_URL}/v1"
        c = BuffyPyClient(ssl=True)
        assert c.api_url == "https://localhost:8008/v1"
        c = BuffyPyClient(url="https://example.de")
        assert c.api_url == "https://example.de:443/v1"

    def ten_meg_static():
        print("## Ten megs static...")
        files = []
        for i in [0, 3]:
            time.sleep(i)
            req_one_mb_cron = bclient.create_request(
                f"{SHITTY_WEBSERVER_BASE_URL}/download/static-content-static-etag?size_bytes=10485760",
                cache_configuration=RequestCacheConfiguration(
                    recaching_strategy=ReCachingStrategy.never()
                ),
            )
            path = f"{STORAGE_CLIENT}/ten_mb_static_never{i}.file"
            req_one_mb_cron.download_response_content_to(path)
            files.append(path)
        # we should recieve 2 times the same cached file
        assert hashfile(files[0]) == hashfile(files[1])

    def ten_meg_din_cron():
        print("## one meg dynamic, CRON strat...")

        files = []
        for i in [0, 16]:
            time.sleep(i)
            req_one_mb_cron = bclient.create_request(
                f"{SHITTY_WEBSERVER_BASE_URL}/download/random-content-static-etag?size_bytes=1048576",
                cache_configuration=RequestCacheConfiguration(
                    recaching_strategy=ReCachingStrategy.cron(
                        cron="* * * * * */10",
                        run_at_start=False,  # recache every 10 seconds
                    )
                ),
            )
            path = f"{STORAGE_CLIENT}/one_dyn_cron_mb{i}.file"
            req_one_mb_cron.download_response_content_to(path)
            files.append(path)

        # the files should differ as it was recached
        assert hashfile(files[0]) != hashfile(files[1])
        # disable recaching as this test is over
        req_one_mb_cron.cache_configuration = RequestCacheConfiguration(
            recaching_strategy=ReCachingStrategy.never()
        )

    def ten_meg_din_age():
        print("## one meg dynamic, AGE strat...")
        files = []
        for i in [0, 1, 16]:
            time.sleep(i)
            req_one_mb_age = bclient.create_request(
                f"{SHITTY_WEBSERVER_BASE_URL}/download/random-content-static-etag?size_bytes=1048575"
            )
            req_one_mb_age.cache_configuration = RequestCacheConfiguration(
                max_cached_versions=3,
                recaching_strategy=ReCachingStrategy.age(
                    seconds=6
                ),  # recache every 6 seconds
            )
            path = f"{STORAGE_CLIENT}/one_dyn_age_mb{i}.file"
            req_one_mb_age.download_response_content_to(path)
            files.append(path)

        # the first 2 files were downloaded before the file 'aged'. they should be the same:
        assert hashfile(files[0]) == hashfile(files[1])
        # the third file was recached by the buffy server after 10 seconds because if its age.
        # it should be a different file
        assert hashfile(files[0]) != hashfile(files[2])
        req_one_mb_age.cache_configuration = RequestCacheConfiguration(
            recaching_strategy=ReCachingStrategy.never()
        )

    def filename_test():

        req_one_mb_age = bclient.create_request(
            f"{SHITTY_WEBSERVER_BASE_URL}/download/random-content-static-etag?size_bytes=1048575"
        )

        req_one_mb_age.cache_configuration = RequestCacheConfiguration(
            recaching_strategy=ReCachingStrategy.never()
        )

        target_path = req_one_mb_age.download_response_content_to(dir=STORAGE_CLIENT)

        assumed_target_path = Path(f"{STORAGE_CLIENT}/rand_bytes.bytes")
        print(target_path, assumed_target_path)
        assert target_path == assumed_target_path

    def test_download_to_dir():

        size = 1048575
        r = bclient.create_request(
            f"{SHITTY_WEBSERVER_BASE_URL}/download/random-content-static-etag?size_bytes={size}"
        )
        p = r.download_response_content_to(dir=STORAGE_CLIENT)
        assert p.stat().st_size == size

    def test_local_download_to_dir():
        c = BuffyPyClient(host="non-existing-host", ssl=False)
        size = 1048576
        r = c.create_request(
            f"{SHITTY_WEBSERVER_BASE_URL}/download/random-content-static-etag?size_bytes={size}"
        )
        p = r.download_response_content_to(dir=STORAGE_CLIENT)
        assert p.stat().st_size == size

    def multiple_requests_with_when_req():
        def start_one_req(index):
            c = BuffyPyClient(url=BUFFY_SERVER_URL, local_download_fallback=False)
            size = 1048573
            r = c.create_request(
                f"{SHITTY_WEBSERVER_BASE_URL}/download/random-content-static-etag?size_bytes={size}",
                cache_configuration=RequestCacheConfiguration(
                    recaching_strategy=ReCachingStrategy.when_requested()
                ),
            )
            p = r.download_response_content_to(f"{STORAGE_CLIENT}/MP_WR_{index}")
            assert p.stat().st_size == size

        req_count = 5
        req_procs: List[multiprocessing.Process] = []
        timeout = 10
        start_time = time.time()
        for i in range(0, req_count):
            req_procs.append(multiprocessing.Process(target=start_one_req, args=(i,)))

        for proc in req_procs:
            proc.start()

        while all([p.is_alive() for p in req_procs]):
            if time.time() - start_time > timeout:
                for proc in req_procs:
                    proc.terminate()
                    raise TimeoutError()
        for proc in req_procs:
            proc.join()
        hashes = []
        for i in range(0, req_count):
            hashes.append(hashfile(f"{STORAGE_CLIENT}/MP_WR_{i}"))
        # TODO: find a way to assert results...
        print(hashes)

    def timed_out_download():
        pass

    def failed_download():
        pass

    # run_test_func(ten_meg_static, 60)

    run_test_func(test_client_params, 5)
    run_test_func(ten_meg_static, 60)
    run_test_func(ten_meg_din_cron, 60)
    run_test_func(ten_meg_din_age, 60)
    run_test_func(filename_test, 10)
    run_test_func(test_download_to_dir, 20)
    run_test_func(test_local_download_to_dir, 20)
    run_test_func(multiple_requests_with_when_req, 20)


except:
    end(exit=False)
    raise
end()



from distutils.command.config import config
import os
import sys
import time
import subprocess
from shittywebserver import run_shitty_webserver
import requests
from pathlib import Path
import json

print("# Create server instances...")
SCRIPT_DIR = "."
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))
STORAGE_BASE_DIR = "./tests/tmp"
STORAGE_CLIENT = f"{STORAGE_BASE_DIR}/client"
STORAGE_SERVER = f"{STORAGE_BASE_DIR}/server"
os.makedirs(STORAGE_CLIENT, exist_ok=True)
os.makedirs(STORAGE_SERVER, exist_ok=True)
os.environ["CONFIGS_STORAGE_BACKEND_CONFIG"] = json.dumps(
    {
        "file_storage_base_path": STORAGE_SERVER,
        "redis_connection_params": {},
    }
)
from buffy.buffyserver.api.v1.models import Request_in, Request, Response
from buffy.buffyserver.main import start_buffy_server
from buffy.buffypyclient import (
    BuffyPyClient,
    RequestCacheConfiguration,
    ReCachingStrategy,
)


BUFFY_BASE_URL = "http://localhost:8008/v1/"
SHITTY_WEBSERVER_BASE_URL = "http://localhost:8088/v1/"


print("Start shittywebserber as test endpoint...")
webserver_proc = run_shitty_webserver(port=8088, run_in_subprocess=True)
time.sleep(1)
print("Start Buffy instance...")
backend_proc, api_proc = start_buffy_server()
time.sleep(1)


def do_request(endpoint: str):
    url = SHITTY_WEBSERVER_BASE_URL + endpoint
    # url = "https://releases.ubuntu.com/22.04/ubuntu-22.04-desktop-amd64.iso"
    r = Request_in(url=url)
    print("RUN:", endpoint, r.json())
    res: requests.Response = requests.put(BUFFY_BASE_URL + "request/", data=r.json())
    res.raise_for_status()
    buffy_req: Request = Request.parse_raw(res.content)
    response_done: bool = False
    time.sleep(60)
    while not response_done:
        rsp_wait = requests.get(
            BUFFY_BASE_URL + f"request/{buffy_req.id}/response/latest", data=r.json()
        )
        if rsp_wait.content == b"null":
            time.sleep(0.5)
            continue
        buffy_res: Response = Response.parse_raw(rsp_wait.content)
        if buffy_res.status in ["wait", "in_progress"]:
            time.sleep(0.5)
            continue
        if buffy_res.status == "ready":
            response_done = True
    print("content_download_url", buffy_res.content_download_path)


# do_request("download/static-content-static-etag?size_bytes=104857600")
# do_request("download/slow?size_bytes=5048500&chunk_wait_time=0.002")
do_request("download/do-nothing")

## clean up

import atexit


def shut_down():
    print("Shut down test env...")
    backend_proc.terminate()
    api_proc.terminate()
    webserver_proc.terminate()


atexit.register(shut_down)



import json
import uuid
import logging
from io import FileIO
from typing import Dict, List, overload, Literal, AsyncGenerator, Union
import redis
from pathlib import Path, PurePath
import hashlib

import asyncio
from buffy.buffyserver.api.v1.models import Response, Request, Request_in
from buffy.tools.utils import url_to_path_dir_name
from buffy.buffyserver.backend.storage.interface import StorageInterface

log = logging.getLogger(__name__)


class RedisStorage(StorageInterface):
    HASHING_ALG = hashlib.md5
    KEY_BASE: str = "bfy"
    KEY_REQUESTS: str = f"{KEY_BASE}:reqs"
    KEY_RESPONSES: str = f"{KEY_BASE}:resp"

    def __init__(
        self,
        file_storage_dir: Path = Path("./buffycache"),
        config={"host": "localhost", "port": 6379},
    ):
        self.db = redis.Redis(**config)
        self.files_base = (
            file_storage_dir
            if isinstance(file_storage_dir, Path)
            else Path(file_storage_dir)
        )
        self.open_file_ios: Dict[Response, FileIO] = {}

    def test_connection(self) -> bool:
        try:
            res = self.db.info()
            if res:
                return True
            else:
                return False
        except redis.exceptions.ConnectionError as e:
            log.error(e)
            log.error("Redis Connection not possible")
            return False

    def _generate_id(self) -> str:
        # as we will work with fairly low quantities, lets shorten the ids.
        return uuid.uuid4().hex[:12]

    def _get_response_content_cache_storage_path(self, response: Response) -> Path:
        request = self.get_request(response.request_id)
        unique_but_readable_subdir = f"{url_to_path_dir_name(request.url)}_{request.id}"
        return Path(
            PurePath(self.files_base, unique_but_readable_subdir, response.version)
        )

    def list_requests(
        self,
        group_name: str = None,  # Request.group_name
    ) -> List[Request]:
        all_req: List[Request] = []
        for key in (
            self.db.keys(f"{self.KEY_REQUESTS}:*")
            if not group_name
            else [f"{self.KEY_REQUESTS}:{group_name}"]
        ):
            group_reqs: List[Request] = [
                Request.parse_raw(raw_req) for raw_req in self.db.hgetall(key).values()
            ]
            all_req.extend(group_reqs)
        all_req = sorted(all_req, key=lambda x: x.inital_request_datetime_utc)
        return all_req

    def get_request(self, request_id: str) -> Request:
        # this is expensive... we need to pick up all requets first. maybe a group_name-ids mapping extra hash can improve this, so we only need to pick up the certain group
        return next(req for req in self.list_requests() if req.id == request_id)

    def create_request(self, request_in: Request_in) -> Request:
        request: Request = Request.from_request(request_in=request_in)
        request.id = self._generate_id()
        request.cache_configuration.request_id = request.id
        db_key = f"{self.KEY_REQUESTS}:{request.group_name if request.group_name else 'DEFAULT_GROUP'}"
        self.db.hset(db_key, request.id, request.json())
        return request

    def update_request(self, request: Request) -> Request:
        db_key = f"{self.KEY_REQUESTS}:{request.group_name if request.group_name else 'DEFAULT_GROUP'}"
        self.db.hset(db_key, request.id, request.json())
        return request

    def list_responses(self, request_id: str) -> List[Response]:
        db_key = f"{self.KEY_RESPONSES}:{request_id}"

        return sorted(
            [
                Response.parse_raw(raw_req)
                for raw_req in self.db.hgetall(db_key).values()
            ],
            key=lambda x: x.version,
        )

    def get_response(self, request_id: str = None, version: str = "latest") -> Response:
        """_summary_

        Args:
            request_id (str, optional) Request.id: _description_. Defaults to None.
            version (Response.version, optional) Response.version: _description_. Defaults to "latest".

        Returns:
            Response: _description_
        """
        if version == "latest":
            all = self.list_responses(request_id)
            return all[-1] if all else None

        db_key = f"{self.KEY_RESPONSES}:{request_id}"
        obj = self.db.hget(db_key, version)
        return Response.parse_raw(obj) if obj else None

    def create_response(self, response: Response) -> Response:
        db_key = f"{self.KEY_RESPONSES}:{response.request_id}"
        response.id = self._generate_id()
        self.db.hset(db_key, response.version, response.json())
        return response

    def update_response(self, response: Response) -> Response:
        db_key = f"{self.KEY_RESPONSES}:{response.request_id}"
        self.db.hset(
            db_key,
            response.version,
            response.json(),
        )
        return response

    def delete_response(self, response: Response):
        db_key = f"{self.KEY_RESPONSES}:{response.request_id}"
        self._get_response_content_cache_storage_path(response).unlink(missing_ok=True)
        self.db.hdel(db_key, response.version)

    ## ↓ CONTENT/FILE HANDLING ↓

    @overload
    async def read_response_content(
        self,
        response: Response,
        as_text: bool = Literal[False],
    ) -> AsyncGenerator[bytes, None]:
        pass

    async def read_response_content(
        self,
        response: Response,
        as_text: bool = Literal[True],
    ) -> AsyncGenerator[str, None]:
        # https://fastapi.tiangolo.com/advanced/custom-response/#streamingresponse
        # https://stackoverflow.com/questions/55873174/how-do-i-return-an-image-in-fastapi/67497103#67497103
        # why async? https://github.com/tiangolo/fastapi/issues/2302
        # file = "/home/tim/Downloads/tails-amd64-4.16.iso"
        file = self._get_response_content_cache_storage_path(response)
        if as_text:
            for line in open(
                file,
                "rt",
            ).readlines():
                yield line
        else:
            with open(
                file,
                "rb",
            ) as f:
                while True:
                    chunk = f.read(1024)
                    if chunk:
                        yield chunk
                    else:
                        break

    def write_response_content(
        self,
        response: Response,
    ) -> FileIO:
        file = self._get_response_content_cache_storage_path(response)
        file.parent.mkdir(parents=True, exist_ok=True)
        return open(file, "ab")

    def get_response_content_hash(self, response: Response, cached_hash: bool = True):
        if response.content_hash_hex and cached_hash:
            return response.content_hash_hex
        req = self.get_request(request_id=response.request_id)
        hash_func = getattr(hashlib, req.validation_hash_type)
        hash_inst = hash_func()

        async def async_wrapper():
            async for chunk in self.read_response_content(
                response=response, as_text=False
            ):

                hash_inst.update(chunk)

        asyncio.run(async_wrapper())
        response.content_hash_hex = hash_inst.hexdigest()
        self.update_response(response=response)
        return response.content_hash_hex

    def _get_response_target_path(self, response: Response):
        pass



import os
import sys
from pathlib import Path, PurePath

SCRIPT_DIR = None
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))

from buffy.tools.stuborn_downloader import (
    StubornDownloader,
    BuffyRequest,
    ResponseDownloadStats,
)

file = "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed22n0003.xml.gz"
file = "https://releases.ubuntu.com/22.04/ubuntu-22.04-desktop-amd64.iso"
file = "https://download-installer.cdn.mozilla.net/pub/firefox/releases/100.0.1/linux-x86_64/de/firefox-100.0.1.tar.bz2"
file = "https://filesamples.com/samples/document/txt/sample3.txt"

print_status = True


def hook_status_updates(status: ResponseDownloadStats):
    if print_status:
        print("STATUS FROM HOOK", status)


def test_target_path():
    trgt = PurePath(SCRIPT_DIR, "tests", "tmp")
    req = BuffyRequest(url=file)
    d = StubornDownloader(request=req)
    # print("STATUS FROM CALLER", d.status)
    d.hook_status_updates = hook_status_updates
    # print(d.get_response_content_attrs(cached_only=False))
    d.download(target_dir=trgt)
    print(d.response_header_fields)
    # print("STATUS FROM CALLER", d.status)
    # print(d.byte_position)
    # print(d.get_response_content_attrs(cached_only=False))


def test_file_obj():
    trgt_dir = PurePath(
        SCRIPT_DIR,
        "tests",
        "tmp",
    )
    trgt_file = PurePath(trgt_dir, "myfile.part")
    file_obj = open(trgt_file, "ab")
    req = BuffyRequest(url=file)
    d = StubornDownloader(request=req)

    d.hook_status_updates = hook_status_updates

    d.download(file_obj=file_obj)
    os.rename(trgt_file, PurePath(trgt_dir, d.get_response_content_attrs().filename))


test_target_path()



from typing import Callable, List, Union, Dict, Type
import time
from fastapi import APIRouter, FastAPI, middleware, Request, Response
from starlette.middleware.base import BaseHTTPMiddleware
import asyncio
from buffy.buffyserver.api.v1.models import (
    Response,
    Request,
    Request_in,
    Request_out,
    RequestCacheConfiguration,
)
from buffy.buffyserver.backend.storage.interface import StorageInterface
from buffy.buffyserver.backend.request_download_service.service import (
    RequestDownloaderService,
)
from fastapi.responses import StreamingResponse
import datetime
from fastapi import status, Response as FastApiResponse

tags_metadata = [
    {
        "name": "Request",
        "description": "Request a remote HTTP resource. Buffy will download this resource in the background and provide it via the `../response/` path",
    },
    {
        "name": "Response",
        "description": "The metadata of a Request outcome. The `../content` of a result can be a json document, a file, a binary file or whatever you requested :)",
        # "externalDocs": {
        #    "description": "External Docs",
        #    "url": "https://fastapi.tiangolo.com/",
        # },
    },
]


def get_v1_router(storage: StorageInterface):
    v1router = APIRouter()
    from fastapi.security import HTTPBearer, HTTPBasicCredentials
    from fastapi import Depends
    import functools

    @v1router.get(
        "/request/",
        tags=["Request"],
        response_model=List[Request_out],
        description="A List all existing requests",
        name="List all existing requests",
    )
    async def list_requests(group_name: str = None) -> List[Request_out]:
        return [
            Request_out.from_request(r)
            for r in storage.list_requests(group_name=group_name)
        ]

    @v1router.put(
        "/request/",
        tags=["Request"],
        response_model=Request_out,
        description="""If the request allready exists it will be returned otherwise a new request will be created.  
        A request is a 'ticket' for waiting and getting the outcoming responses metadata and content.  
        For existing Requests you can, if applicable, access earlier cached versions of the responses and accelerate your client code if freshness of the content is not your priority 🚀""",
        name="Find or create a request",
        status_code=status.HTTP_202_ACCEPTED,
        responses={201: {"model": Request_out}},
    )
    async def find_or_create_responses_request(
        request_in: Request_in, api_response: FastApiResponse
    ) -> Request_out:
        """
        matching_req = next(
            (
                req
                for req in storage.list_requests(group_name=request_in.group_name)
                if req.signature == request_in.signature
            ),
            None,
        )
        """
        ts = time.time()
        matching_req: Request = None
        for req in storage.list_requests(group_name=request_in.group_name):
            if req.signature == request_in.signature:
                matching_req = req
                break
        if matching_req:
            matching_req.latest_request_datetime_utc = datetime.datetime.utcnow()
            matching_req.latest_requests.append(ts)
            storage.update_request(matching_req)
            out_req = Request_out.from_request(matching_req)
            out_req.request_timestamp = ts
            return out_req
        api_response.status_code = status.HTTP_201_CREATED

        new_req = storage.create_request(request_in)
        new_req.latest_requests = [ts]
        new_req.inital_request_datetime_utc = datetime.datetime.utcnow()
        new_req = storage.update_request(new_req)
        out_req = Request_out.from_request(new_req)
        out_req.request_timestamp = ts
        return out_req

    @v1router.get(
        "/request/{request_id}",
        tags=["Request"],
        response_model=Request,
        description="""If you allready know the __id__ of your Request, you can explicitly demand a certain Request here""",
        name="Get a certain request",
    )
    async def get_request(request_id: str) -> Request_out:
        return storage.get_request(Request_out.from_request(request_id))

    @v1router.get(
        "/request/{request_id}/status/{request_timestamp}",
        tags=["Request"],
        status_code=status.HTTP_200_OK,
        responses={102: {}},
        description="Check if your request is registered by the backend. You can now try to query a Response object at the `/v1/response/*` endpoints. \
            **Hint**: This does not indicate that your response content is ready/available.",
    )
    async def get_response_by_request_timestamp(
        request_id: str, request_timestamp: float, api_response: FastApiResponse
    ):
        if not (
            request_timestamp
            in storage.get_request(request_id=request_id).latest_requests
        ):
            return {"status": "ok"}
        api_response.status_code = status.HTTP_102_PROCESSING
        return None

    @v1router.get(
        "/request/{request_id}/cache-config",
        tags=["Request"],
        response_model=RequestCacheConfiguration,
        description="""Get the (re)cache configuration of this request""",
        name="Get a certain requests (re)cache config",
    )
    async def get_request_cache_config(request_id: str) -> RequestCacheConfiguration:
        return storage.get_request(request_id).cache_configuration

    @v1router.put(
        "/request/{request_id}/cache-config",
        tags=["Request"],
        response_model=RequestCacheConfiguration,
        description="""Get the (re)cache configuration of this request""",
        name="Set or Update a certain requests (re)cache config",
    )
    async def updsert_request_cache_config(
        request_id: str,
        request_cache_config: RequestCacheConfiguration,
    ) -> Request:
        request: Request = storage.get_request(request_id)
        # update the config requets id, if the config was copied from another request
        request_cache_config.request_id = request_id
        request.cache_configuration = request_cache_config
        request = storage.update_request(request)
        return request.cache_configuration

    @v1router.get(
        "/request/{request_id}/response",
        tags=["Response"],
        response_model=List[Response],
        description="""Get a list of all cached/ongoing/failed Responses of a Request.""",
        name="List all responses of a request",
    )
    async def list_request_results(request_id: str):
        return storage.list_responses(request_id)

    @v1router.get(
        "/request/{request_id}/response/t/{tag}",
        tags=["Response"],
        response_model=List[Response],
    )
    async def list_responses_by_tag(request_id: str, tag: str):
        all_responses = storage.list_responses(request_id)
        return [res for res in all_responses if res.tags and tag in res.tags]

    @v1router.get(
        "/request/{request_id}/response/v/{version}",
        tags=["Response"],
        response_model=Response,
    )
    async def get_response(request_id: str, version: str = "latest"):
        return storage.get_response(request_id=request_id, version=version)

    @v1router.get(
        "/request/{request_id}/response/v/{version}/content",
        tags=["Response"],
        response_class=StreamingResponse,
        responses={
            200: {
                "content": {"any": {}},
                "description": "Return whatever the source response media type is.",
            }
        },
    )
    async def serve_responses_content(request_id: str, version: str):
        response = storage.get_response(request_id=request_id, version=version)
        is_text_like: bool = False
        if response.content_attributes.media_type:
            is_text_like = response.content_attributes.media_type in [
                "text",
                "json",
                "application/json",
            ] or response.content_attributes.media_type.startswith("text/")

        headers: Dict = {}
        if (
            response.content_attributes.content_size_bytes
            and response.content_attributes.content_size_bytes > 0
        ):
            headers["Content-Length"] = str(
                response.content_attributes.content_size_bytes
            )
        if response.content_attributes.content_disposition:
            headers[
                "Content-Disposition"
            ] = response.content_attributes.content_disposition
        return StreamingResponse(
            storage.read_response_content(response=response, as_text=is_text_like),
            headers=headers,
            media_type=response.content_attributes.media_type,
        )

    @v1router.put(
        "/request/{request_id}/response/v/{version}/pin",
        tags=["Response"],
        response_model=Response,
        description="""When a Response is marked as `valid`-ated an older Response version can be purged out of the cache.  
        If you pin a certain version, it will be keept no matter what.""",
        name="Pin a certain Response version in the cache",
    )
    async def update_response(request_id: str, version: str, value: bool):
        response = storage.get_response(request_id=request_id, version=version)
        if value in [True, "true", "True", 1, "1"]:
            response.pinned = True
        else:
            response.pinned = False
        return storage.update_response(response)

    @v1router.put(
        "/request/{request_id}/response/v/{version}/valid",
        tags=["Response"],
        response_model=Response,
        description="""This endpoint is intented for client side soundness validation. You can mark a file as corrupt or the opposite. it will not be offered anymore if not valid (except its explicitly requested).  
        Do not confuse "soundness","file corruption" validation with "the file is in the correct scheme for my purpose". For this case we have the `/request/{request_id}/response/v/{version}/operative/{tag}` endpoint.
    """,
        name="Mark a certain Response version in the cache as not corrupt",
    )
    async def update_response(request_id: str, version: str, value: bool):
        response = storage.get_response(request_id=request_id, version=version)
        if value in [True, "true", "True", 1, "1"]:
            response.valid = True
        else:
            response.valid = False
        return storage.update_response(response)

    @v1router.put(
        "/request/{request_id}/response/v/{version}/{tag}",
        tags=["Response"],
        response_model=Response,
        description="""Here you can tag a certain response version.  
        The `tag` could be your clients version.
        This way you can make your client, dependend on external resources, having a stable reproducibe state. as it will always get the same file served.  
        Which is the whole purpose of this venture :) 
    """,
        name="Mark a certain response version. E.g. as 'worked-for-me'",
    )
    async def update_response(request_id: str, version: str, tag: str):
        response = storage.get_response(request_id=request_id, version=version)
        if not response.tags:
            response.tags = []
        response.tags.append(tag)
        return storage.update_response(response)

    return v1router



import time
import inspect
from typing import Dict, Optional, Literal, Union, Type, List, TYPE_CHECKING, TypedDict
from pydantic import BaseModel, Field
import datetime
from croniter import croniter
from logging import getLogger

log = getLogger(__name__)
if TYPE_CHECKING:
    from buffy.buffyserver.api.v1.models import Request, Response
    from buffy.buffyserver.backend.storage.interface import StorageInterface


class BaseReCachingStrategy(BaseModel):
    strategy_name: str = None

    def __eq__(self, other):
        if inspect.isclass(other):
            return other.__name__ == self.__class__.__name__
        elif isinstance(other, str):
            return other == self.__class__.__name__
        return isinstance(other, self.__class__)

    def request_needs_recache(
        self,
        request: "Request",
        latest_response: "Response",
        strategy_state: dict,
        storage: "StorageInterface",
    ) -> bool:
        pass


class ReCachingStrategy:
    """
    Wrapper class to contain all possible ReCachingStrategies.
    A ReCachingStrategy defines how an when a remote resource should be updated
    """

    @classmethod
    def list(cls) -> List[Type[BaseReCachingStrategy]]:
        """List all available strategies"""
        return [
            cls_attribute
            for cls_attribute in cls.__dict__.values()
            if inspect.isclass(cls_attribute)
        ]

    class never(BaseReCachingStrategy):
        """never -
        Once downlaoded we will serve only this cached file. no matter how often it is requested.
        This is usefull for files that will be 100% static.
        """

        strategy_name: Literal["never"] = Field(default="never", const=True)

        def request_needs_recache(
            self,
            request: "Request",
            latest_response: "Response",
            strategy_state: dict,
            storage: "StorageInterface",
        ) -> bool:
            return False

    class age(BaseReCachingStrategy):
        """Age Strategy
        
        Buffy will redownload when the cached version has a certain age in seconds.  
        The default age is 3600sec.  
        This is usefull for dynamic remote resources, where the webserver wont provide any informations about the \
            state of the resource like "etag", "last_modified_datetime" or "size_in_bytes"

        Args:
            seconds (int): \
                default:3600 - The duration a resource will be cached until it gets a redownload from the source \
                
        """

        strategy_name: Literal["age"] = Field(default="age", const=True)
        seconds: int = 3600

        def request_needs_recache(
            self,
            request: "Request",
            latest_response: "Response",
            strategy_state: dict,
            storage: "StorageInterface",
        ) -> bool:
            if request.latest_requests and latest_response.cached_datetime_utc is None:
                return True
            if (
                latest_response.cached_datetime_utc is not None
                and (
                    datetime.datetime.utcnow() - latest_response.cached_datetime_utc
                ).total_seconds()
                > request.cache_configuration.recaching_strategy.seconds
            ):
                return True
            return False

    class cron(BaseReCachingStrategy):
        """cron - Strategy
        Buffy will redownload the remote resource in an certain interval. Defined in the Linux cron format.
        Similar to the "age"-Strategy, but allows more flexibility for more complex situations.
        This is usefull for dynamic remote resources, where the webserver wont provide any informations about the state of the resource like "etag", "last_modified_datetime" or "size_in_bytes"

        params:
            cron - str - a standard cron definition https://en.wikipedia.org/wiki/Cron

            run_at_start - bool - regardless of the cron schedule, Buffy will redownload the resource once when Buffy just started
        """

        strategy_name: Literal["cron"] = Field(default="cron", const=True)
        cron: str = "0 0 * * *"
        run_at_start: bool = True

        def request_needs_recache(
            self,
            request: "Request",
            latest_response: "Response",
            strategy_state: dict,
            storage: "StorageInterface",
        ) -> bool:
            if TYPE_CHECKING:
                request.cache_configuration.recaching_strategy: ReCachingStrategy.cron = (
                    request.cache_configuration.recaching_strategy
                )

            if request.id not in strategy_state:
                #  init cron
                start_time: datetime.datetime = datetime.datetime.utcnow()
                strategy_state[request.id]: Dict[
                    str, Union[croniter, datetime.datetime]
                ] = {}
                new_cron_job = croniter(
                    request.cache_configuration.recaching_strategy.cron, start_time
                )
                strategy_state[request.id]["cron_job"] = new_cron_job
                strategy_state[request.id]["start_time"] = start_time
                strategy_state[request.id]["next"] = new_cron_job.get_next(
                    datetime.datetime
                )
                if request.cache_configuration.recaching_strategy.run_at_start:
                    return True
            next_cron_time: datetime.datetime = strategy_state[request.id]["next"]
            cron_job: croniter = strategy_state[request.id]["cron_job"]
            if next_cron_time <= datetime.datetime.utcnow():
                strategy_state[request.id]["next"] = cron_job.get_next(
                    datetime.datetime
                )
                return True
            return False

    class when_requested(BaseReCachingStrategy):
        """when_requested  - Strategy
        Default Strategy!
        Buffy will redownload the resource when the client requests the resource.
        This is the most simple scenario which is closest to a classic just-download case but with a fallback safeguard;
        If the remote resource fails to download, buffy can fallback to an older cached version of the resource, if available.

        **HINT**: Multiple request during one server tick (default 0.3s), will be bundled and share one response. \
            Example: If you simultaneously request 5 numbers from a REST API that generates random numbers, you will most likely just get same number for every request.
            In cases like this you maybe better off, bypassing a caching middleware like Buffy.
        """

        strategy_name: Literal["when_requested"] = Field(
            default="when_requested", const=True
        )

        def request_needs_recache(
            self,
            request: "Request",
            latest_response: "Response",
            strategy_state: dict,
            storage: "StorageInterface",
        ) -> bool:
            if request.latest_requests:
                return True
            return False

    class when_remote_changed(BaseReCachingStrategy):
        """when_remote_changed  - Strategy
        Buffy will ask the webserver of the remote resource about the state of the resource.
        Via "etag", "last_modified_datetime" or "size_in_bytes" Buffy can determine if the resource changed.
        If it has changed Buffy will download and cache the newer version.
        This is usefull for scenarios where the client need to have the newest version instant and reliable available.

        params:
            check_interval_sec
        """

        strategy_name: Literal["when_remote_changed"] = Field(
            default="when_remote_changed", const=True
        )
        check_interval_sec: int = 43200

        def request_needs_recache(
            self,
            request: "Request",
            latest_response: "Response",
            strategy_state: dict,
            storage: "StorageInterface",
        ) -> bool:
            if TYPE_CHECKING:
                request.cache_configuration.recaching_strategy: ReCachingStrategy.when_remote_changed = (
                    request.cache_configuration.recaching_strategy
                )
            from buffy.tools.stuborn_downloader import StubornDownloader

            if not request.id in strategy_state:
                strategy_state[request.id] = 0
            remote_file_change_last_check_time_unix_epoch: float = strategy_state[
                request.id
            ]
            if (
                remote_file_change_last_check_time_unix_epoch
                + request.cache_configuration.recaching_strategy.check_interval_sec
                > time.time()
            ):
                return False
            remote_file_change_last_check_time_unix_epoch = time.time()
            strategy_state[request.id] = remote_file_change_last_check_time_unix_epoch
            potential_next_response_header = StubornDownloader(
                request=request
            ).get_response_content_attrs()
            if (
                not potential_next_response_header
                and not potential_next_response_header.etag
                and not potential_next_response_header.last_modified_datetime_utc
                and not potential_next_response_header.size_in_bytes
            ):
                err_msg = f"The file can not be recached with strategy 'ReCachingStrategy.{ReCachingStrategy.when_remote_changed.__name__}'! Can not determine if target file of request (id:'{request.id}',url:'{request.url}' changed. The server at '{request.url}' does not provide any informations. Please use another caching strategy than {ReCachingStrategy.when_remote_changed}."
                log.error(err_msg)
                latest_response.download_stats.error = err_msg
                storage.update_response(latest_response)
                return False

            attr_comparison_sequence: List[str] = [
                "etag",
                "last_modified_datetime",
                "size_in_bytes",
            ]
            for attr in attr_comparison_sequence:
                if (
                    hasattr(potential_next_response_header, attr)
                    and hasattr(latest_response, attr)
                ) and (
                    getattr(potential_next_response_header, attr)
                    != getattr(latest_response, attr)
                ):
                    return True
            return False



from typing import List, Dict, Tuple, Literal, Set
import threading
from io import FileIO
import traceback
from buffy.buffyserver.api.v1.models import Request, Response, Request_in
from buffy.tools.stuborn_downloader import StubornDownloader
from buffy.tools.utils import extract_domain


class Download:
    class ThreadWithExpectionStackTraceStatus(threading.Thread):
        def __init__(self, *args, **kwargs):
            self.error_stack_trace: str = None
            self.error: Exception = None
            self.failed: bool = False
            super().__init__(**kwargs)

        def run(self):
            try:
                self._target(*self._args, **self._kwargs)
            except Exception as err:
                self.error = err
                self.error_stack_trace = traceback.format_exc()
                self.failed = True
            finally:
                # Avoid a refcycle if the thread is running a function with
                # an argument that has a member that points to the thread.
                del self._target, self._args, self._kwargs

    def __init__(self, request: Request, response: Response, file_target: FileIO):
        self.request = request
        self.response = response
        self.domain = extract_domain(request.url)
        self.downloader = StubornDownloader(request=request)
        self.thread = Download.ThreadWithExpectionStackTraceStatus(
            target=self.downloader.download,
            args=(
                None,
                file_target,
            ),
        )

    @property
    def status(self) -> Literal["wait", "run", "finished"]:
        if self.thread.ident is None:
            return "wait"
        elif self.thread.is_alive():
            return "run"
        else:
            return "finished"


class RequestDownloadQueue:
    def __init__(self):
        self.downloads: List[Download] = []

    def get_downloads(
        self,
        by_domain: str = None,
        by_status: Literal["wait", "run", "finished"] = None,
    ) -> List[Download]:
        return [
            d
            for d in self.downloads
            if (d.domain == by_domain or by_domain is None)
            and (d.status == by_status or by_status is None)
        ]

    @property
    def domains(self) -> Set:
        return set([d.domain for d in self.downloads])

    def add_download(self, request: Request, response: Response, file_target: FileIO):
        self.downloads.append(
            Download(request=request, response=response, file_target=file_target)
        )

    def remove_download(self, download: Download):
        self.downloads.remove(download)



from typing import Dict, Literal, Union
from Configs import ConfigBase


class DEFAULT(ConfigBase):
    LOG_LEVEL: str = "INFO"

    #####################
    ### SERVER PARAMS ##
    ###################

    # UVICORN_PARAMS - json/dict - Configuration of the uvicorn webserver which is publishing the Buffy Server API
    # You put any `uvicorn.run` parameter in this dict.
    # See https://www.uvicorn.org/settings/ for all parameters
    UVICORN_PARAMS: Dict = {"port": 8008, "host": "0.0.0.0"}

    #####################
    ### SERVER STORAGE PARAMS ##
    ###################

    STORAGE_BACKEND_MODULE: Literal["redis"] = "redis"

    # STORAGE_BACKEND_CONFIG - json/dict - Configuration for the storage module used.
    # atm only the `redis`-storage-module is supported.
    # The value for key `redis_connection_params` can contain any parameter for the redis python client https://redis.readthedocs.io/en/latest/connections.html#generic-client
    # example:
    # STORAGE_BACKEND_CONFIG = {"file_storage_base_path":"./","redis_connection_params":{"host":"145.23.45.23","port":6379,"password":"s3cret"}}
    STORAGE_BACKEND_CONFIG: Dict = {}

    # STORAGE_BACKEND_FILE_DIR - str: the storage backend can cache files to the local file system. This directory path defines the location for these files
    STORAGE_BACKEND_FILE_DIR: str = "./buffy-server-cache"

    # STORAGE_CLEAN_EVERY_N_SEC - int - How often should be checked if there are old responses that are not needed anymore.
    STORAGE_CLEAN_EVERY_N_SEC: int = 60

    # STORAGE_VALIDATE_EVERY_N_SEC - int - When validating every cached response is checked for unexpexted changes. If any occure buffy will delete these responses.
    # This can happen on unrealiable storage or when the storage is whipped by external processes (e.g. container restart without volumes)
    STORAGE_VALIDATE_EVERY_N_SEC: int = 43200

    # STORAGE_VALIDATE_ON_BOOT - bool - see STORAGE_VALIDATE_EVERY_N_SEC but should be one run started when buffy server is starting?
    STORAGE_VALIDATE_ON_BOOT: bool = True

    # DOWNLOAD_SERVICE_MAX_DOWNLOADS - int - How many downloads simultaniously should be started in the background?
    DOWNLOAD_SERVICE_MAX_DOWNLOADS: int = 4
    # DOWNLOAD_SERVICE_MAX_DOWNLOADS_PER_DOMAIN - int - How many downloads simultaniously should be started in the background per domain?
    DOWNLOAD_SERVICE_MAX_DOWNLOADS_PER_DOMAIN: int = 1



import os
import sys
from fastapi import FastAPI
from typing import Dict, Tuple, Type
from multiprocessing import Process
import logging
from getversion import get_module_version
import time
import click
import redis
import json

log = logging.getLogger(__name__)
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "../..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))

from Configs import getConfig
from buffy.buffyserver.config import DEFAULT

from buffy.buffyserver.api.v1.api import (
    get_v1_router,
    tags_metadata,
)
from buffy.buffyserver.backend.storage.interface import StorageInterface
from buffy.buffyserver.backend.storage.redis_storage import RedisStorage
from buffy.buffyserver.backend.request_download_service.service import (
    RequestDownloaderService,
)
from pathlib import Path, PurePath

config_file_path = PurePath(Path(__file__).parent.resolve(), "config.py")

logging.basicConfig(
    level=getConfig(config_classes_pathes=[config_file_path]).LOG_LEVEL,
    format="%(asctime)-15s %(processName)-8s %(module)-8s %(levelname)-8s:  %(message)s",
    handlers=[logging.StreamHandler((sys.stdout))],
)
# atm we only have one storage module (redis)
storage_module_mapping: Dict[str, Type[StorageInterface]] = {"redis": RedisStorage}


def _start_api_service():
    import uvicorn

    config: DEFAULT = getConfig(config_classes_pathes=[config_file_path])
    # config: DEFAULT = getConfig()

    app = FastAPI(openapi_tags=tags_metadata)
    storage: StorageInterface = storage_module_mapping[config.STORAGE_BACKEND_MODULE](
        file_storage_dir=config.STORAGE_BACKEND_FILE_DIR,
        config=config.STORAGE_BACKEND_CONFIG,
    )
    app.include_router(get_v1_router(storage), prefix="/v1")

    uvicorn.run(app, **config.UVICORN_PARAMS)


def _start_backend_service():
    # config: DEFAULT = getConfig()
    config: DEFAULT = getConfig(config_classes_pathes=[config_file_path])
    storage: StorageInterface = storage_module_mapping[config.STORAGE_BACKEND_MODULE](
        file_storage_dir=config.STORAGE_BACKEND_FILE_DIR,
        config=config.STORAGE_BACKEND_CONFIG,
    )
    service = RequestDownloaderService(
        storage=storage,
        max_simultaneous_downloads=config.DOWNLOAD_SERVICE_MAX_DOWNLOADS,
        max_simultaneous_downloads_per_domain=config.DOWNLOAD_SERVICE_MAX_DOWNLOADS_PER_DOMAIN,
    )
    service.validate_storage_every_n_sec = config.STORAGE_VALIDATE_EVERY_N_SEC
    service.clean_storage_every_n_sec = config.STORAGE_CLEAN_EVERY_N_SEC
    service.validate_storage_at_boot = config.STORAGE_VALIDATE_ON_BOOT
    service.run()


def start_buffy_server(
    env_vars: Dict = None, watchdog: bool = True
) -> Tuple[Process, Process]:
    if env_vars:
        for k, v in env_vars.items():
            os.environ[k] = json.dumps(v) if isinstance(v, dict) else v

    import buffy
    import pydantic
    import uvicorn

    config: DEFAULT = getConfig(config_classes_pathes=[config_file_path])
    log.info(f"BuffyServer version: '{get_module_version(buffy)[0]}'")
    log.info(f"Pydantic version: '{get_module_version(pydantic)[0]}'")
    log.info(f"Uvicorn version: '{get_module_version(uvicorn)[0]}'")
    log.info(
        f"Storage backend config ('STORAGE_BACKEND_CONFIG'): '{config.STORAGE_BACKEND_CONFIG}'"
    )
    log.info(
        f"Storage backend file dir ('STORAGE_BACKEND_FILE_DIR'): '{config.STORAGE_BACKEND_FILE_DIR}'"
    )
    log.info(f"Uvicorn config ('UVICORN_PARAMS'): '{config.UVICORN_PARAMS}'")
    # toDo: make a more stuborn watchdog process starter.
    backend = Process(target=_start_backend_service, name="Buffy_backend")
    api = Process(target=_start_api_service, name="Buffy_api_service")
    backend.start()
    api.start()
    if watchdog:
        # todo try n-time to rescue/restart a service
        healthy: bool = True
        while healthy:
            if not backend.is_alive():
                backend.join()
                api.terminate()
                api.join()
                raise ValueError(f"Backend exited with code {backend.exitcode}")
            if not api.is_alive():
                api.join()
                backend.terminate()
                backend.join()
                raise ValueError(f"API Service exited with code {api.exitcode}")
            time.sleep(1)
    return backend, api


@click.command(name="buffy-server")
@click.option(
    "--debug/--no-debug",
    help="Enable debug logging",
    default=False,
)
@click.option(
    "--redis-url",
    "-r",
    help="Url to a running Redis database. e.g. 'redis://localhost:6379/0'. For details on the format see https://redis.readthedocs.io/en/latest/connections.html#redis.Redis.from_url",
)
@click.option(
    "--storage-location",
    "-s",
    help="Buffy server will cache http request into this location. e.g. `/var/lib/buffy`",
)
def start_buffy_server_cli(debug: bool, redis_url: str, storage_location: str):
    """Start the buffy server"""
    if debug:
        click.echo(f"Debug mode is on")
        log.setLevel("DEBUG")
        log.debug(f"redis_url({type(redis_url)}):'{redis_url}'")
        log.debug(f"storage_location({type(storage_location)}):'{storage_location}'")
    server_env_vars = {}
    if storage_location:
        Path(storage_location).mkdir(exist_ok=True)
        server_env_vars["CONFIGS_STORAGE_BACKEND_FILE_DIR"] = storage_location
    if redis_url:
        r = redis.Redis.from_url(redis_url)
        server_env_vars["CONFIGS_STORAGE_BACKEND_CONFIG"] = r.get_connection_kwargs()
    start_buffy_server(env_vars=server_env_vars)


if __name__ == "__main__":

    start_buffy_server()



import requests

remote_file_url = "https://filesamples.com/samples/document/txt/sample3.txt"
response: requests.Response = requests.head(remote_file_url)
response_header_fields = response.headers
print(response_header_fields)



import hashlib
import time
import typing
import inspect
from typing import Dict, Optional, Literal, Union, Type, List
from pydantic import BaseModel, Field
import datetime
import enum
from buffy.buffyserver.api.v1.models_recaching_strategies import ReCachingStrategy
import json

SUPPORTED_HASH_TYPES = typing.Literal[tuple(hashlib.algorithms_guaranteed)]
AVAILABLE_RECHACHING_STRATEGIES = typing.Union[tuple(ReCachingStrategy.list())]
# print("AVAILABLE_RECHACHING_STRATEGIES", AVAILABLE_RECHACHING_STRATEGIES)
# RECACHING_STRATEGIES = typing.Literal[tuple(ReCachingStrategy.str_list())]


class RequestCacheConfiguration(BaseModel):
    """The request cache configuration can be attached to a request to give Buffy server instruction how and when to cache a request.
    At the heart of the configuration is a [ReCachingStrategy](BuffyReCachingStrategies). the other option is how many versions of a request-response should be cached.

    **HINT**: This is pydantic model. Every `class-attribute` is also a `parameter`

    **example usage**
    ```python
    import json
    from buffy.buffypyclient import BuffyPyClient
    c = BuffyPyClient()

    # Define a strategy. Buffy server has to re-cache the target url every 3601 seconds.
    strategy = ReCachingStrategy.age(seconds=3601)
    # Lets pack the strategy in a RequestCacheConfiguration config
    config = RequestCacheConfiguration(recaching_strategy=strategy, max_cached_versions=3)
    # Lets attach the config to our request
    req = c.create_request("https://wikipedia.org",cache_configuration=config)
    ```
    """

    request_id: Optional[str]

    # recaching_strategy: Union[ReCachingStrategy.never, ReCachingStrategy.age] = Field(
    #    discriminator="strategy_name"
    # )
    # d = {"propertyName": "strategy_name"}
    recaching_strategy: AVAILABLE_RECHACHING_STRATEGIES = Field(
        default=ReCachingStrategy.never(),
        discriminator="strategy_name",
    )
    """A [strategy](/BuffyReCachingStrategies) instructs buffy how to cache request-responses. (e.b. by age of the cached file, or regulary by a cron job,...)"""
    max_cached_versions: int = 2
    """When more than `max_cached_versions` request-responses are cached by the Buffy server, any older cached version will be deleted."""

    # To be implemented later # unpack_response: bool = False


class Request_in(BaseModel):
    url: str
    http_method: Literal["get", "post", "put"] = "get"
    http_header_fields: Optional[Dict] = {}

    # https://docs.python-requests.org/en/latest/user/quickstart/#passing-parameters-in-urls
    http_query_params: Optional[Dict] = {}
    http_request_body: Optional[Dict] = {}
    group_name: Optional[str] = "DEFAULT_GROUP"
    description: Optional[str] = None
    documentation_link: Optional[str] = None
    validation_hash_type: Optional[SUPPORTED_HASH_TYPES] = "md5"
    cache_configuration: Optional[RequestCacheConfiguration] = None

    @property
    def signature(self) -> int:
        return hash(
            (
                self.url,
                self.http_method,
                json.dumps(self.http_header_fields),
                json.dumps(self.http_query_params),
                json.dumps(self.http_request_body),
                self.group_name,
            )
        )


class Request_out(Request_in):
    id: Optional[str]
    cache_configuration: RequestCacheConfiguration
    request_timestamp: float = None

    @classmethod
    def from_request(cls, request_in: Request_in) -> "Request":
        if request_in.cache_configuration is None:
            request_in.cache_configuration = RequestCacheConfiguration()
        return cls(
            **{
                k: getattr(request_in, k)
                for k in cls.__fields__.keys()
                if k in request_in.__fields__.keys()
            }
        )

        # return cls(**request_in.dict())


class Request(Request_out):
    inital_request_datetime_utc: datetime.datetime = None
    latest_request_datetime_utc: Optional[datetime.datetime] = None
    latest_requests: List[float] = [time.time()]


class ResponseContentAttributes(BaseModel):
    media_type: Optional[str] = None
    etag: Optional[str] = None
    content_size_bytes: Optional[int] = None
    last_modified_datetime_utc: Optional[datetime.datetime] = None
    filename: Optional[str] = None
    content_disposition: Optional[str] = None

    def is_empty(self):
        return not all(dict(self).values())


class ResponseDownloadStats(BaseModel):
    """Statistics about the running or finished download of remote HTTP resource

    **HINT**: This is pydantic model. Every `class-attribute` is also a `parameter`
    """

    downloaded_bytes: int = 0
    """How many bytes are allready downloaded"""
    avg_bytes_per_sec: int = None
    """The average download speed."""
    download_running_duration_sec: float = None
    """How long is the download or running or how long did it run when finished."""
    download_start_time_unix_ts: float = None
    """The moment when the Buffy server started the download"""
    total_bytes_to_download: int = None
    """The size of the response content. This value can initaly be unrealiable if the remote webserver reported a wrong value, but will be realiable correct when the download is finished."""
    state: Literal["init", "downloading", "failed", "finished"]
    """State of the download. if `failed` there will be Stacktrace in `error`"""
    error: str = None
    """Stacktrace if download failed"""


class Response(BaseModel):
    """Metadata object for a request response.

    **HINT**: This is pydantic model. Every `class-attribute` is also a `parameter`
    """

    id: str = None
    request_id: str
    version: str
    """A compact date string, declarating the versions name"""
    previous_version: Optional[str] = None
    """If there is a predating version, you can find the version name here"""
    next_version: Optional[str] = None
    """If there is a newer version, you can find the version name here"""
    status: Literal["wait", "in_progress", "ready", "failed"]
    """State of the reponses download on Buffy server side"""
    content_download_path: Optional[str] = None
    """Path to downlaod the responses content from the Buffy server API. No base domain included!"""
    cached_datetime_utc: Optional[datetime.datetime] = None
    """Moment when Buffy server finished the download"""
    request_datetime_utc: Optional[datetime.datetime] = None
    """Moment when a client requested the file. This is not appliable to all strategies."""
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition
    pinned: Optional[bool] = False
    """!WIP!: A pinned response will not be deleted by the garbage collector if `Request.cache_configuration.max_cached_versions` is reached."""
    tags: Optional[List[str]] = None
    """!EXPERIMENTAL!: Clients can add tags to a reponse to manage multiple versions"""
    valid: Optional[bool] = None
    """!OBSOLETE!: Will be replaced by `tags`"""
    download_stats: ResponseDownloadStats = None
    """Statistics about the downlaod by Buffy server"""
    content_attributes: ResponseContentAttributes = None
    """Attributes the download could gather about the content. mostly based on HTTP headers"""
    content_hash_hex: str = None
    """A hex hash of the response content. The hash alg is defined in `Request.validation_hash_type`"""



from __future__ import annotations
from io import FileIO
import io
from urllib.parse import urlparse
import re
from typing import List, Literal, Type, Union
import hashlib
from pathlib import Path


def extract_domain(url: str):
    parsed = urlparse(url)
    return parsed.hostname


def url_to_path_dir_name(
    url: str,
    object_id: str = None,
    max_length: int = 64,
):
    max_length_minus_id_length = max_length - (len(object_id) + 1 if object_id else 0)
    parsed = urlparse(url)
    base_url_components = f"{parsed.hostname}{'-'+str(parsed.port) if parsed.port and str(parsed.port) not in ['80','443'] else ''}".lower()
    base_url_components += f"{'-'+parsed.path if parsed.path else ''}{'-'+parsed.params if parsed.params else ''}{'-'+parsed.query if parsed.query else ''}".lower()
    cleaned_base = re.sub("[^0-9a-zA-Z]+", "-", base_url_components)
    return f"{cleaned_base[0:max_length_minus_id_length]}{'-'+object_id if object_id else ''}".replace(
        "--", "-"
    )


def is_file_like(obj):
    return (
        isinstance(obj, io.TextIOBase)
        or isinstance(obj, io.BufferedIOBase)
        or isinstance(obj, io.RawIOBase)
        or isinstance(obj, io.IOBase)
        or issubclass(type(obj), io.TextIOBase)
        or issubclass(type(obj), io.BufferedIOBase)
        or issubclass(type(obj), io.RawIOBase)
        or issubclass(type(obj), io.IOBase)
    )


def hashfile(file: Union[str, Path, FileIO], alg: Type[hashlib._Hash] = None) -> str:
    """_summary_

    Args:
        file (Union[str, Path, FileIO]): A path or file like object
        alg (Type[hashlib._Hash], optional): The hashing alg to create the hash of the file. Defaults to hashlib.sha256.

    Returns:
        str: A hex str representing the hash of the file
    """
    # source: https://www.geeksforgeeks.org/compare-two-files-using-hashing-in-python/

    # A arbitrary (but fixed) buffer
    # size (change accordingly)
    # 65536 = 65536 bytes = 64 kilobytes
    if alg == None:
        alg = hashlib.sha256

    BUF_SIZE = 65536

    # Initializing the sha256() method
    hash_: hashlib._Hash = alg()
    seek_pos: int = None
    if is_file_like(file):
        seek_pos = file.tell()
        file.seek(0)
        f = file
    else:
        f = open(file, "rb")
    while True:

        # reading data = BUF_SIZE from
        # the file and saving it in a
        # variable
        data = f.read(BUF_SIZE)

        # True if eof = 1
        if not data:
            break

        # Passing that data to that sh256 hash
        # function (updating the function with
        # that data)
        hash_.update(data)

    if seek_pos:
        # we had an opened file obj. lets restore it to the old state
        file.seek(seek_pos)
    else:
        # we created a local file obj based on a path. lets close it
        f.close()

    # sha256.hexdigest() hashes all the input
    # data passed to the sha256() via sha256.update()
    # Acts as a finalize method, after which
    # all the input data gets hashed hexdigest()
    # hashes the data, and returns the output
    # in hexadecimal format
    return hash_.hexdigest()


class HttpHeaderContentDispositionParser:
    def __init__(self, content_disposition: str):
        self.raw: str = content_disposition
        self.type: Literal["inline", "attachment", "form-data"] = None
        self.filename_unsafe: str = None
        self.name: str = None
        self._parse()

    def _parse(self):
        if self.raw:
            segments: List[str] = self.raw.split(";")
            if len(segments) in [1, 2, 3]:
                for seg in segments:
                    self._parse_segment(seg)
            else:
                raise ValueError(
                    f"Invalid http header 'content-disposition' syntax. expected sth. as in https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition#syntax got: '{self.raw}'"
                )

    def _parse_segment(self, segment: str):
        segment = segment.strip()
        if segment.lower() in ["inline", "attachment", "form-data"]:
            self.type = segment.lower()
        elif segment.lower().startswith("filename="):
            self.filename_unsafe = (
                segment.split("=")[1].replace('"', "").replace("'", "")
            )
        elif segment.lower().startswith("name="):
            self.name = segment.split("=")[1].replace('"', "").replace("'", "")

    @property
    def filename(self):
        keepcharacters = (" ", ".", "_", "(", ")", "-")
        return "".join(
            c for c in self.filename_unsafe if c.isalnum() or c in keepcharacters
        ).rstrip()



from functools import cache
from genericpath import isfile
from http import client
import io
import ssl
from typing import Dict, Literal, List, Optional, Union, Iterator, Annotated
import logging
from xml.dom import ValidationErr
from buffy.buffyserver.api.v1.models import (
    Request as ApiRequest,
    Request_in as ApiRequest_in,
    Response as ApiResponse,
    ResponseDownloadStats,
)
from buffy.buffyserver.api.v1.models_recaching_strategies import ReCachingStrategy
from buffy.buffyserver.api.v1.models import RequestCacheConfiguration

from pathlib import Path, PurePath
from buffy.tools.stuborn_downloader import StubornDownloader
import requests
from io import FileIO
import tempfile
import threading
import time
from pydantic import parse_obj_as, parse_raw_as
from requests.auth import HTTPBasicAuth
from urllib.parse import urlparse

log = logging.getLogger(__name__)


class BuffyServerError(Exception):
    pass


class BuffyCacheEmptyError(Exception):
    pass


class BuffyPyClient:
    """Python client to abstract communication with the Buffy server REST API"""

    buffy_server_timeout_sec = 5

    class Request:
        """Return object of [`Buffy.BuffyPyClient.create_request()`](/BuffyPyClient/#buffy.buffypyclient.buffypyclient.BuffyPyClient.create_request)

        Args:
            force_version (str, optional): The size of byte chunks you want to iterate. Defaults to 64 kibibyte.

        """

        def __init__(self, request: ApiRequest_in, client: "BuffyPyClient"):
            self._client = client
            self._api_request_in: ApiRequest_in = request
            self._api_request: ApiRequest = None
            self._api_response: ApiResponse = None
            self._use_local_download: bool = False
            self._local_downloader: StubornDownloader = None
            self._local_temp_file: FileIO = None
            self.force_version: str = None
            """if you want to have a certain version of all cached responses, set `force_version`. To list all available versions use """
            self.prefere_version: str = None
            self.prefere_tag: str = None
            """If a response version with this tag is available it will be preferred to be served to the client."""
            self.prefere_pinned: bool = False
            self.prefere_cached_instead_of_waiting: bool = False
            self.fallback_to_older_version = True
            self.buffy_server_http_headers: Dict = {}
            """Header fields send to the Buffy server API. This can be used e.g. if Buffy is behind a auth reverse proxy"""
            self.buffy_server_http_auth: HTTPBasicAuth = None
            """HTTP Auth fields send to the Buffy server API. This can be used e.g. if Buffy is behind a auth reverse proxy"""

        @property
        def cache_configuration(self) -> RequestCacheConfiguration:
            if not self._api_request:
                return self._api_request_in.cache_configuration
            else:
                return self._api_request.cache_configuration

        @cache_configuration.setter
        def cache_configuration(self, cache_configuration: RequestCacheConfiguration):
            if not self._api_request:
                self._api_request_in.cache_configuration = cache_configuration
            else:
                self._api_request.cache_configuration = cache_configuration
                try:
                    res = requests.put(
                        f"{self._client.api_url}/request/{self._api_request.id}/cache-config",
                        data=cache_configuration.json()
                        if cache_configuration
                        else None,
                        timeout=self._client.buffy_server_timeout_sec,
                        headers=self.buffy_server_http_headers,
                        auth=self.buffy_server_http_auth,
                    )
                    res.raise_for_status()
                except (
                    requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                ) as e:
                    log.warning(
                        f"Can not reach BuffyServer @ '{self._client.api_url}'. Cache configuration is not updated on server side."
                    )

        def order(self):
            """Send the request to the Buffy server.
            When calling [`Buffy.BuffyPyClient.create_request()`](/BuffyPyClient/#buffy.buffypyclient.buffypyclient.BuffyPyClient.create_request) `order` will be called automaticly.
            If you want to call `order`manually, set `create_request(hold_request_order=True)`

            """
            try:
                res: requests.Response = requests.put(
                    f"{self._client.api_url}/request/",
                    data=self._api_request_in.json(),
                    timeout=self._client.buffy_server_timeout_sec,
                    verify=False,
                    headers=self._client.http_headers,
                    auth=self._client.http_auth,
                )
                res.raise_for_status()
            except requests.exceptions.ConnectionError as e:
                if self._client.local_download_fallback:
                    log.warning(
                        f"Can not reach BuffyServer @ '{self._client.api_url}'. Fallback to local uncached download."
                    )
                    self._use_local_download = True
                    self.download_thread = threading.Thread(target=self._local_download)
                    self.download_thread.start()
                    return
                else:
                    log.error(f"Can not reach BuffyServer @ '{self._client.api_url}'.")
                    raise e
            except:
                log.error(f"Request data: {self._api_request_in.json()}")
                if hasattr(res, "content"):
                    log.error(f"Response content {res.content}")
                raise

            self._api_request: ApiRequest = ApiRequest.parse_raw(res.content)
            if (
                self._api_request_in.cache_configuration
                and self._api_request_in.cache_configuration
                != self._api_request.cache_configuration
            ):
                # the caller updated the cache config. lets update on server side
                self.cache_configuration = self._api_request_in.cache_configuration
            self._wait_for_request_registration()

        def _is_request_registered(self) -> bool:
            res: requests.Response = requests.get(
                f"{self._client.api_url}/request/{self._api_request.id}/status/{self._api_request.request_timestamp}",
                headers=self.buffy_server_http_headers,
                auth=self.buffy_server_http_auth,
            )
            res.raise_for_status()

            if res.status_code == 102:
                return False
            elif res.status_code == 200:
                return True

        def _wait_for_request_registration(self):
            """Wait until the request is processed by the backend and a possible response had arrived in the database"""
            if self._api_request is None:
                raise ValidationErr(
                    f"Request for '{self._api_request_in.url}' must be started with method 'BuffyPyClient.Request.order() first'"
                )
            while not self._is_request_registered():
                time.sleep(0.3)

        def _wait_for_response_completion(
            self, render_progressbar: bool = False, timeout_sec: int = None
        ):
            if timeout_sec or render_progressbar:
                raise NotImplementedError()
            is_complete: bool = False
            while not is_complete:
                stats: ResponseDownloadStats = self._get_download_stats()
                if stats.state == "finished":
                    return
                if stats.state == "failed":
                    raise BuffyServerError(stats.error)
                time.sleep(0.3)

        def _wait_for_response_content_attributes(self, timeout_sec: int = None):
            if timeout_sec:
                raise NotImplementedError()
            if self._use_local_download:
                self._local_downloader.get_response_content_attrs()
            else:
                is_complete: bool = False
                while not is_complete:
                    resp = self._find_response()

                    if resp.content_attributes.filename is not None:
                        return
                    time.sleep(0.3)

        def _find_response(self, override_prefere_version: str = None) -> ApiResponse:
            # Check if we already called _find_response() and found a response
            if self._api_response:
                # update state of reponse
                self._get_download_stats(response=self._api_response)
                if self._api_response.status != "failed":
                    # if the state of the reponse is healthy we can serve it.
                    # otherwise we continue the function and need to find another one
                    return self._api_response
            prefer_version = (
                override_prefere_version
                if override_prefere_version
                else self.prefere_version
            )
            if self._use_local_download:
                return None

            elif (
                not (
                    prefer_version
                    or self.prefere_tag
                    or self.prefere_pinned
                    or self.prefere_cached_instead_of_waiting
                )
                and self._api_request is None
            ):
                raise ValidationErr(
                    f"Request for '{self._api_request_in.url}' must be started with method 'BuffyPyClient.Request.order() first'"
                )

            response: ApiResponse = None
            if prefer_version:
                # The client pefers a certain version.

                response_url: str = f"{self._client.api_url}/request/{self._api_request.id}/v/{prefer_version}"
                response_raw = requests.get(
                    response_url,
                    headers=self.buffy_server_http_headers,
                    auth=self.buffy_server_http_auth,
                ).content
                if not response_raw and self.force_version:
                    raise ValueError(
                        f"Can not get BuffyResponse at '{response_url}'. Probably version '{prefer_version}' set in BuffyClient.Request.prefere_version is not existent"
                    )
                response = ApiResponse.parse_raw(response_raw)
                if (
                    not self.prefere_cached_instead_of_waiting
                    and response.status != "failed"
                    or self.force_version
                ):
                    # no matter if this response is still downloading, the client wants this version
                    return response
            if not prefer_version or not response:
                all_responses_raw = requests.Response = requests.get(
                    f"{self._client.api_url}/request/{self._api_request.id}/response",
                    headers=self.buffy_server_http_headers,
                    auth=self.buffy_server_http_auth,
                ).content
                buffy_responses: List[ApiResponse] = parse_raw_as(
                    List[ApiResponse], all_responses_raw
                )
                for resp in buffy_responses:
                    if resp.tags and self.prefere_tag in resp.tags:
                        response = resp
                    if self.prefere_pinned and resp.pinned:
                        response = resp
                if not response and buffy_responses:
                    # we just take the latest response
                    response = buffy_responses[-1]
            if (
                self.prefere_cached_instead_of_waiting
                and response.status in ["wait", "in_progress"]
                and response.previous_version
            ) or response.status == "failed":
                if response.previous_version:
                    response = self._find_response(
                        override_prefere_version=response.previous_version
                    )
                elif response.status == "failed":
                    raise BuffyCacheEmptyError(
                        f"Current response for request '{self._api_request.url}' failed to download and no previous version is cached. sorry!"
                    )
            self._api_response = response
            return response

        def download_response_content(self, chunk_size: int = 65536) -> Iterator[bytes]:
            """Stream/Iterate through the response content delivered by the Buffy server.

            **example use**
            ```python
            from buffy.buffypyclient import BuffyPyClient
            c = BuffyPyClient()
            req = c.create_request(
                url="https://meowfacts.herokuapp.com/"
            )
            cat_fact = ""
            for chunk in req.download_response_content():
                cat_fact += chunk.decode("utf-8")
            print(json.loads(cat_fact))
            ```

            Args:
            Args:
                chunk_size (int, optional): The size of byte chunks you want to iterate. Defaults to 64 kibibyte.

                Iterator[bytes]: The reponse content in byte chunks
            """
            self._wait_for_response_completion()
            if not self._use_local_download:
                response = self._find_response()
                download_url = (
                    f"{self._client.api_url}/{response.content_download_path}"
                )
                with requests.get(
                    download_url,
                    headers=self.buffy_server_http_headers,
                    auth=self.buffy_server_http_auth,
                    stream=True,
                ) as r:
                    r.raise_for_status()
                    for chunk in r.iter_content(chunk_size=chunk_size):
                        yield chunk
            else:
                self._local_temp_file.seek(0)
                while True:
                    chunk = self._local_temp_file.read(chunk_size)
                    if not chunk:
                        break
                    yield chunk

        def get_filename(self, fallback_filename="download") -> Union[str, None]:
            """Try to evaluate the filename of the response content.
            This is based on HTTP headers like \
                [`Content-Disposition`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition) and \
                [`Content-Type`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type)

            Args:
                fallback_filename (str, optional): A filename to fallback if `get_filename` can not evaluate a filename. Defaults to "download".

            Returns:
                The filename as string or `None` if no filename cloud be evaluated and `fallback_filename` is set to `None`
            """
            self._wait_for_response_content_attributes()

            if self._use_local_download:
                filename = self._local_downloader.get_response_content_attrs(
                    cached_only=True
                ).filename
            else:
                filename = self._find_response().content_attributes.filename
            return filename if filename else fallback_filename

        def download_response_content_to(
            self, target: Union[str, Path, FileIO] = None, dir: Union[str, Path] = None
        ) -> Path:
            """Download the response content delivered by the Buffy server.

            **example use**
            ```python
            from buffy.buffypyclient import BuffyPyClient
            c = BuffyPyClient()
            req = c.create_request(url="https://www.foaas.com/awesome/:tim")
            content_path = req.download_response_content_to(dir="/tmp")
            print(content_path)
            # > `/tmp/download.html`
            ```

            Args:
                target (Union[str, Path, FileIO], optional): Download target as a file like object \
                    or a file path.
                dir (Union[str, Path], optional): Alternative to `target`. \
                    A directory as download path. Buffy tries to evaluate the filename \ 
                    but will fallback to `download`. Defaults to None.

            Returns:
                Path: The final path of the downloaded content.
            """
            file = target
            if dir and target:
                raise ValueError("Use target or dir as filesink not both")
            if dir:
                target = Path(PurePath(dir, self.get_filename()))
            if isinstance(target, str):
                target = Path(target)
            if isinstance(target, Path):
                file = open(target, "wb")
            for chunk in self.download_response_content():
                file.write(chunk)
            if isinstance(target, (str, Path)):
                # We only close the file object if we created ourselves
                file.close()
            return target

        def _local_download(self):
            # in extra thread option?
            self._local_downloader = StubornDownloader(self._api_request_in)
            self._local_downloader.CLOSE_FILE_OBJ = False
            self._local_temp_file = tempfile.TemporaryFile()
            self._local_downloader.download(file_obj=self._local_temp_file)

        def _wait_for_local_download(self):
            raise NotImplementedError()

        def _get_download_stats(
            self, response: ApiResponse = None
        ) -> ResponseDownloadStats:
            if self._use_local_download:
                return self._local_downloader.status
            else:
                # update response status
                response_url: str = f"{self._client.api_url}/request/{self._api_request.id}/response/v/{response.version if response else self._find_response().version}"
                response_raw = requests.get(
                    response_url,
                    headers=self.buffy_server_http_headers,
                    auth=self.buffy_server_http_auth,
                ).content
                self._api_response = ApiResponse.parse_raw(response_raw)
                return self._api_response.download_stats

        def list_cached_versions(self) -> List[ApiResponse]:
            """Get a list of all cached responses of the request

            ```python
            from buffy.buffypyclient import BuffyPyClient
            c = BuffyPyClient()
            req = c.create_request(url="https://www.foaas.com/awesome/:tim")
            cached_responses = req.list_cached_versions()
            for resp in cached_responses:
                print(resp.id)
            ```

            Returns:
                List[ApiResponse]: _description_
            """
            response_url: str = (
                f"{self._client.api_url}/request/{self._api_request.id}/response/"
            )
            response_raw = requests.get(
                response_url,
                headers=self.buffy_server_http_headers,
                auth=self.buffy_server_http_auth,
            ).content
            return parse_raw_as(List[ApiResponse], response_raw)

        def __del__(self):
            if self._local_temp_file and not self._local_temp_file.closed:
                self._local_temp_file.close()

    def __init__(
        self,
        url: str = None,
        host: str = None,
        port: int = 8008,
        base_path: str = None,
        group_name: str = None,
        api_version: Literal["v1"] = "v1",
        ssl: bool = False,
        http_auth: HTTPBasicAuth = None,
        http_headers: Dict = {},
        local_download_fallback: bool = True,
    ):
        self.http_auth = None
        if not host and not url:
            host = "localhost"
        if host:
            self.ssl = ssl
            self.host = host
            self.port = port
            self.base_path = base_path
        elif url:
            parsed_url = urlparse(url)
            self.ssl = True if parsed_url.scheme in ["https", None] else False
            self.host = parsed_url.hostname
            if not parsed_url.port and not self.ssl:
                self.port = 80
            elif not parsed_url.port and self.ssl:
                self.port = 443
            elif parsed_url.port:
                self.port = parsed_url.port

            self.base_path = parsed_url.path
            if parsed_url.username:
                self.http_auth = HTTPBasicAuth(
                    username=parsed_url.username, password=parsed_url.password
                )

        if http_auth:
            self.http_auth = http_auth
        self.http_headers = http_headers
        self.group_name = group_name
        self.api_version = api_version
        self.local_download_fallback = local_download_fallback

    @property
    def api_url(self):
        api_url = f"http{'s' if self.ssl else ''}://{self.host}:{self.port}/{self.base_path + '/' if self.base_path else ''}{self.api_version}"
        log.debug(f"api_url: {api_url}")
        return api_url

    def create_request(
        self,
        url: str,
        http_method: Literal["get", "post", "put"] = "get",
        http_query_params: Dict = {},
        http_request_body: Optional[Dict] = {},
        http_header_fields: Dict = {},
        info_description: str = None,
        info_link: str = None,
        hold_request_order: bool = False,
        cache_configuration: RequestCacheConfiguration = None,
    ) -> Request:
        """
        Define a HTTP request to a remote resource.


        **example usage**
        ```python
        import json
        from buffy.buffypyclient import BuffyPyClient
        c = BuffyPyClient() 
        req = c.create_request("https://wikipedia.org")
        ```

        Args:
            url (str): The full URL to the remote resource. Example: "https://mydomain.org/myfile.txt" \
                Optional: Query params can be supplied via a Dict to `http_query_params`
            http_method (Literal["get", "post", "put"], optional): [HTTP Request method](https://en.wikipedia.org/wiki/Hypertext_Transfer_Protocol#Request_methods) for the to be created request
            http_query_params (Dict, optional): [HTTP query string](https://en.wikipedia.org/wiki/Query_string)
            http_request_body (Optional[Dict], optional): [HTTP message body](https://en.wikipedia.org/wiki/HTTP_message_body)
            http_header_fields (Dict, optional): [HTTP header fields](https://en.wikipedia.org/wiki/List_of_HTTP_header_fields)
            info_description (str, optional): This is a documentation string for your request. Buffy server will save your request. \
                For later traceability it can help to give your request some metadata.
            info_link (str, optional): Same as `info_description` but for external metadata
            hold_request_order (bool, optional): The Buffy client will send the request to the buffy server instantly. \
                If for any reasons you need to delay the request, set `hold_request_order` to `True`
            cache_configuration (RequestCacheConfiguration, optional): Buffy server can be configured on when to (re-)download  your request. \
                An instance of `RequestCacheConfiguration` is needed

        Returns:
            BuffyPyClient.Request: An object to handle your request. this is needed to download the content of your request
        """
        api_req = ApiRequest_in(url=url)
        api_req.http_method = http_method
        api_req.http_query_params = http_query_params
        api_req.http_request_body = http_request_body
        api_req.http_header_fields = http_header_fields
        api_req.description = info_description
        api_req.documentation_link = info_link
        api_req.cache_configuration = cache_configuration

        req = BuffyPyClient.Request(api_req, self)
        if hold_request_order:
            return req
        req.order()
        return req



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="DZDBuffy",
    description="Versioning Caching kinda-Proxy for decoupling external responses",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.connect.dzd-ev.de/dzdpythonmodules/buffy",
    author="Tim Bleimehl",
    author_email="tim.bleimehl@helmholtz-muenchen.de",
    license="MIT",
    packages=["buffy", "buffy.buffypyclient", "buffy.buffyserver"],
    install_requires=[
        "fastapi",
        "uvicorn",
        "pydantic",
        "requests",
        "redis",
        "croniter",
        "DZDConfigs",
        # "tqdm",
        "getversion",
        "Click",
    ],
    extras_require={
        "test": ["shittywebserver"],
        "docs": [
            "mkdocs",
            "mkdocstrings[python]",
            "mkdocs-autorefs",
            "mkdocs-material",
            "mkdocs-render-swagger-plugin",
        ],
    },
    python_requires=">=3.9",
    zip_safe=False,
    include_package_data=True,
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        # "local_scheme": "node-and-timestamp"
        "local_scheme": "no-local-version",
        "write_to": "version.py",
    },
    setup_requires=["setuptools_scm"],
    entry_points={
        "console_scripts": [
            "buffy-server=buffy.buffyserver.main:start_buffy_server_cli",
        ],
    },
)



import logging
from buffy.buffyserver.backend.storage.interface import StorageInterface
from buffy.buffyserver.api.v1.models import Request, Response

log = logging.getLogger(__name__)


class StorageJanitor:
    def __init__(self, storage: StorageInterface):
        self.storage = storage

    def cleanup_storage(self):
        """Remove responses that are too old according to `Request.cache_configuration.max_cached_versions`"""
        for req in self.storage.list_requests():
            self._clean_up_request_responses(req)

    def _clean_up_request_responses(self, request: Request):
        ready_responses = [
            res
            for res in self.storage.list_responses(request_id=request.id)
            if res.status in ["ready"]
        ]
        if request.cache_configuration.max_cached_versions < len(ready_responses):
            for resp in ready_responses[
                request.cache_configuration.max_cached_versions :
            ]:
                self.storage.delete_response(resp)
            ready_responses = ready_responses[
                : request.cache_configuration.max_cached_versions
            ]

    def verifiy_storage(self):
        """Remove responses that have issues with their content like unexpected changes or just missing
        This can be a very expensive operation. Use wisely!
        """
        for req in self.storage.list_requests():
            self._verifiy_request_responses(req)

    def _verifiy_request_responses(self, request: Request):
        for resp in [
            res
            for res in self.storage.list_responses(request_id=request.id)
            if res.status in ["ready"]
        ]:
            old_hash = resp.content_hash_hex
            try:
                current_hash = self.storage.get_response_content_hash(
                    resp, cached_hash=False
                )
            except FileNotFoundError:
                # file is missing. we have a garbage response here. lets just delete it
                log.warning(
                    f"Request.id: '{request.id}' Response.id: '{resp.id}' -> Error during validation hashing. Response will be deleted"
                )
                self.storage.delete_response(resp)
                # go to next response
                continue
            if old_hash != current_hash:
                # the file changed. this means from our perspective the file is corrupt
                log.warning(
                    f"Request.id: '{request.id}' Response.id: '{resp.id}' -> Response content changed unexpected. Response will be deleted"
                )
                self.storage.delete_response(resp)



import os
from typing import Dict, Callable, Union
from io import FileIO, BytesIO
import requests
from requests.structures import CaseInsensitiveDict
from email.utils import parsedate_to_datetime
from pathlib import Path, PurePath
from urllib.parse import urlparse
import mimetypes
import time
import logging
import hashlib

# rename to avoid confusions with requests.Response and requests.Request

from buffy.buffyserver.api.v1.models import (
    Request_in as BuffyRequest,
    ResponseContentAttributes,
    ResponseDownloadStats,
)
from buffy.tools.utils import HttpHeaderContentDispositionParser

log = logging.getLogger(__name__)


class StubornDownloader:
    FILE_EXTENSION_FOR_PARTIAL_DOWNLOADED_FILE: str = ".part"
    HTTP_METHOD_MAP: Dict[str, Callable] = {
        "get": requests.get,
        "put": requests.put,
        "post": requests.post,
    }
    # https://stackoverflow.com/questions/23369625/ideal-chunk-size-for-python-requests
    CHUNK_SIZE: int = 65536
    DEFAULT_WAIT_TIME_SEC_ON_429: int = 2

    MAX_DOWNLOAD_ATTEMPTS: int = 6
    CLOSE_FILE_OBJ: bool = True
    # important status codes
    # 429 - Too Many Requests - https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/429
    #   Solution: A "Retry-After" header might be included to this response indicating how long to wait before making a new request.
    def __init__(self, request: "BuffyRequest"):
        self.request = request
        self.byte_position = 0
        self.overwrite_filename = None
        self.fallback_filename = "download"
        self.call_report_hook_every_n_chunks_downloaded = 10
        self.accept_ranges: bool = None
        self.response_header_fields: CaseInsensitiveDict = None
        self.status: ResponseDownloadStats = None
        self.finished: bool = False
        self.hash: str = None

        self._status_updates(
            ResponseDownloadStats(
                downloaded_bytes=0,
                avg_bytes_per_sec=None,
                download_running_duration_sec=None,
                download_start_time_unix_ts=None,
                total_bytes_to_download=None,
                state="init",
            )
        )

    def hook_status_updates(self, status: ResponseDownloadStats):
        # mb_per_sec: float = round(status.avg_bytes_per_sec / 1048576, 2)
        pass

    def get_response_content_attrs(
        self, cached_only: bool = False
    ) -> ResponseContentAttributes:
        if not self.response_header_fields and not cached_only:
            response: requests.Response = None
            try:
                if self.request.http_method == "get":
                    response: requests.Response = requests.head(
                        self.request.url,
                        params=self.request.http_query_params,
                        headers=self.request.http_header_fields,
                        timeout=1,
                    )
                    self.response_header_fields = response.headers
                    response.raise_for_status()
                else:
                    raise
            except:
                # webserver does not support "head" calls or we have no 'get'. Then just pretend a download
                try:
                    response: requests.Response = self.HTTP_METHOD_MAP[
                        self.request.http_method
                    ](
                        self.request.url,
                        params=self.request.http_query_params,
                        headers=self.request.http_header_fields,
                        stream=True,
                        timeout=1,
                    )
                    self.response_header_fields = response.headers
                    response.close()
                except (
                    requests.exceptions.ReadTimeout,
                    requests.exceptions.ConnectionError,
                ):
                    if response:
                        response.close()
                    return ResponseContentAttributes()
        elif not self.response_header_fields and cached_only:
            return ResponseContentAttributes()
        response_content_attrs = ResponseContentAttributes(
            etag=self.response_header_fields.get("ETag", None),
            size_in_bytes=self.response_header_fields.get("Content-Length", None),
            filename=self._guess_filename(
                headers=self.response_header_fields,
                url=self.request.url,
                fallback_filename=None,
            ),
            content_disposition=self.response_header_fields.get("Content-Disposition"),
        )
        _media_type = self.response_header_fields.get("Content-Type", None)
        if _media_type:
            response_content_attrs.media_type = _media_type.split(";")[0]
        _last_modified = self.response_header_fields.get("Last-Modified", None)
        if _last_modified:
            response_content_attrs.last_modified_datetime_utc = parsedate_to_datetime(
                _last_modified
            )
        response_content_attrs = self._correct_content_length(response_content_attrs)
        return response_content_attrs

    def _correct_content_length(
        self, response_content_attrs: ResponseContentAttributes
    ) -> ResponseContentAttributes:
        if self.finished:
            if response_content_attrs.content_size_bytes is not None and (
                response_content_attrs.content_size_bytes
                != self.status.downloaded_bytes
            ):
                log.warning(
                    f"""'Content-length' for file '{self.request.url}' was announced as '{response_content_attrs.content_size_bytes}' bytes,  but actually is {self.status.downloaded_bytes} bytes.
                    Buffy downloader will correct this information for the clients response header."""
                )
            response_content_attrs.content_size_bytes = self.status.downloaded_bytes
        return response_content_attrs

    def download(
        self,
        target_dir: Union[Path, str] = None,
        file_obj: Union[FileIO, BytesIO] = None,
    ) -> Union[PurePath, FileIO]:
        if (target_dir and file_obj) or (not target_dir and not file_obj):
            err_msg = f"Expected 'target_dir' OR 'file_obj' for storing download. Got both or none for remote file '{self.request.url}'.\n'target_dir'='{target_dir}'\n'file_obj'={file_obj}\n"
            log.error(err_msg)
            raise ValueError(err_msg)
        elif file_obj and isinstance(file_obj, FileIO):
            if file_obj.mode != "ab":
                err_msg = f"Expected file_obj to be in mode 'ab'. got {file_obj.mode}"
                raise ValueError(err_msg)
        result: Union[PurePath, FileIO] = None
        download_attempts: int = 0
        while not self.finished:
            download_attempts += 1
            try:
                if self.request.validation_hash_type:
                    self._hash_func = getattr(
                        hashlib, self.request.validation_hash_type
                    )()
                else:
                    self._hash_func

                result = self._download(
                    target_dir=target_dir,
                    file_obj=file_obj,
                    start_at_byte=self.byte_position if self.accept_ranges else 0,
                )
                self.finished = True
                self.status.state = "finished"
                self._status_updates(self.status)
                if self._hash_func:
                    self.hash = self._hash_func.hexdigest()

            except requests.exceptions.RequestException as e:
                # we only want to catch http/network/requests errors. other errors are propaply not solved by retrying
                if download_attempts >= self.MAX_DOWNLOAD_ATTEMPTS:
                    self.status.state = "failed"
                    self._status_updates(self.status)
                    raise e
        return result

    def _download(
        self,
        target_dir: Union[Path, str],
        file_obj: FileIO,
        start_at_byte: int = 0,
    ) -> PurePath:

        request_header_fields = CaseInsensitiveDict(self.request.http_header_fields)
        response_complete: bool = False

        while not response_complete:
            response: requests.Response = self.HTTP_METHOD_MAP[
                self.request.http_method
            ](
                self.request.url,
                params=self.request.http_query_params,
                headers=request_header_fields,
                stream=True,
                timeout=60,
            )
            response_complete = True
            self.accept_ranges = (
                True
                if response.headers.get("Accept-Ranges", False) == "bytes"
                else False
            )
            if response.status_code == 429:
                # 429 Too Many Requests # wait for xxx seconds
                retry_after_sec = int(response.headers.get("Retry-After", None))
                retry_after_sec = (
                    self.DEFAULT_WAIT_TIME_SEC_ON_429
                    if not retry_after_sec
                    else retry_after_sec
                )
                log.warning(f"Too many requests. wait {retry_after_sec} seconds...")
                response_complete = False
                time.sleep(int(retry_after_sec))

            elif not self.accept_ranges and start_at_byte != 0:
                raise ValueError(
                    f"'Accept-Ranges' is false but 'start_at_byte' required by caller. Server does not provide resuming downloads or partial downloads."
                )
            elif (
                self.accept_ranges
                and start_at_byte != 0
                and request_header_fields.get("Range", None) is None
            ):
                # this is a resumed download. We request to not download the whole file but start a specific byte position via the header field Range
                # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Range
                request_header_fields["Range"] = f"bytes={start_at_byte}"
                response_complete = False

        response.raise_for_status()
        self.response_header_fields = response.headers
        total_length_bytes = response.headers.get("Content-Length", None)
        if self.overwrite_filename:
            file_name = self.overwrite_filename
        else:
            file_name = self._guess_filename(
                headers=response.headers,
                url=self.request.url,
                fallback_filename=self.fallback_filename,
            )

        if not file_obj:
            # caller wants us to manage file target
            if not target_dir:
                target_dir = "./"
            target_path = PurePath(target_dir, file_name)

            temp_target_path = PurePath(
                str(target_path) + self.FILE_EXTENSION_FOR_PARTIAL_DOWNLOADED_FILE
            )
            write_target = open(temp_target_path, "ab")
        else:
            write_target = file_obj

        if start_at_byte == 0:
            # reset any previous download attempts
            write_target.seek(0)
            write_target.truncate()
        i: int = 0
        start_time = time.time()
        # START DOWNLOAD
        self._status_updates(
            ResponseDownloadStats(
                downloaded_bytes=self.byte_position,
                avg_bytes_per_sec=0,
                download_running_duration_sec=0,
                download_start_time_unix_ts=start_time,
                total_bytes_to_download=total_length_bytes,
                state="downloading",
            )
        )
        for chunk in response.iter_content(chunk_size=self.CHUNK_SIZE):
            if chunk:
                write_target.write(chunk)
                self.byte_position += len(chunk)
                if self._hash_func:
                    self._hash_func.update(chunk)
                if i % self.call_report_hook_every_n_chunks_downloaded == 0:
                    current_time = time.time()
                    self._status_updates(
                        ResponseDownloadStats(
                            downloaded_bytes=self.byte_position,
                            avg_bytes_per_sec=self.byte_position
                            / (current_time - start_time),
                            download_running_duration_sec=current_time - start_time,
                            download_start_time_unix_ts=start_time,
                            total_bytes_to_download=total_length_bytes,
                            state="downloading",
                        )
                    )

            i += 1
        current_time = time.time()
        self._status_updates(
            ResponseDownloadStats(
                downloaded_bytes=self.byte_position,
                avg_bytes_per_sec=self.byte_position / (current_time - start_time),
                download_running_duration_sec=current_time - start_time,
                download_start_time_unix_ts=start_time,
                total_bytes_to_download=total_length_bytes,
                state="downloading",
            )
        )
        response.close()
        if (file_obj and self.CLOSE_FILE_OBJ) or not file_obj:
            write_target.close()
        if not file_obj:
            # lets move our temp content to callers desired target
            try:
                os.remove(target_path)
            except FileNotFoundError:
                pass
            os.rename(temp_target_path, target_path)
        return target_path if not file_obj else file_obj

    def _guess_filename(
        self, headers: CaseInsensitiveDict, url: str, fallback_filename: str = None
    ) -> str:
        # FILENAME VIA CD
        # try to get filename the offical way via content-disposition
        try:
            content_disposition = HttpHeaderContentDispositionParser(
                headers.get("content-disposition")
            )
        except ValueError as err:
            content_disposition = None
            log.error(err)
        if content_disposition and content_disposition.filename_unsafe:
            return content_disposition.filename
        # if content_disposition and content_disposition.type in ["inline", "form-data"]:
        #    return None

        # ALTERNATIVES METHODS
        ## FILE EXTENSION
        ## first lets try to get the file extension
        ext: str = None
        content_type: str = headers.get("Content-Type", None)
        if content_type:
            content_type = content_type.split(";")[0]
        if content_type:
            # try to get it via Content-Type header
            ext = mimetypes.guess_extension(content_type)
            ext = ext.replace(".", "") if ext else None
        ## FILENAME VIA URL
        ## try to pick it out of the url
        filename = (
            PurePath(urlparse(url).path).name
            if PurePath(urlparse(url).path).suffix
            else None
        )
        if filename:
            return filename
        ## fallback to default filename with a last try to guess the filename type and at least generate a correct filename extension
        if fallback_filename and ext:
            return fallback_filename + "." + ext
        if ext:
            return self.fallback_filename + "." + ext
        return self.fallback_filename

    def _status_updates(self, status: ResponseDownloadStats):
        # mb_per_sec: float = round(avg_speed_bytes_per_sec / 1048576, 2)
        self.status = status
        # we copy the object as its not intended to mutate the status from extern
        self.hook_status_updates(ResponseDownloadStats(**dict(status)))



import os
import sys

SCRIPT_DIR = "."
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))

from buffy.buffypyclient import (
    BuffyPyClient,
    ReCachingStrategy,
    RequestCacheConfiguration,
)

import json

# connect to buffy server
from buffy.buffypyclient import (
    BuffyPyClient,
    ReCachingStrategy,
    RequestCacheConfiguration,
)

# connect to buffy server
c = BuffyPyClient(host="localhost", port=8008, ssl=False)

wikidata_query = """
SELECT DISTINCT ?Country ?CountryLabel ?gov_form ?gov_formLabel WHERE {
    ?Country p:P31/ps:P31 wd:Q6256;
        wdt:P37 ?official_language ;
        wdt:P297 ?ISO_3166_1_alpha_2_code ;
    SERVICE wikibase:label { bd:serviceParam wikibase:language "en". }
    OPTIONAL { ?Country wdt:P47 ?sharesBorderWith. }
    OPTIONAL { ?Country wdt:P122 ?gov_form. }
    } ORDER BY ?CountryLabel
"""

# define a strategy on server background  recaching. Countries are relatively static concepts.
# In this example we are happy, if we update the request data every quarter.
# This will relieve the wikidata.org server, no matter how often we request the data.
# Also we wont get on any blacklist if we request to often.
strategy = ReCachingStrategy.age(seconds=60 * 60 * 24 * 92)
# Lets pack the strategy in our Cache config
config = RequestCacheConfiguration(recaching_strategy=strategy, max_cached_versions=3)

# create a request
req = c.create_request(
    url="https://query.wikidata.org/sparql",
    http_query_params={"format": "json", "query": wikidata_query},
    http_header_fields={
        "user-agent": "MyTestClient",
        "Accept": "application/json",
    },
    cache_configuration=config,
)
# save requested file
countries = ""
for chunk in req.download_response_content():
    countries += chunk.decode("utf-8")
print(json.loads(countries))



import os
import sys

SCRIPT_DIR = "."
if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))

from buffy.buffypyclient import (
    BuffyPyClient,
    ReCachingStrategy,
    RequestCacheConfiguration,
)

remote_file_url = "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed22n0003.xml.gz"
remote_file_url = "https://www.dundeecity.gov.uk/sites/default/files/publications/civic_renewal_forms.zip"
remote_file_url = "https://filesamples.com/samples/document/txt/sample3.txt"

# connect to buffy server
c = BuffyPyClient(host="localhost", ssl=False)
c.buffy_server_timeout_sec = 1
# create a request

# optional: define a re-caching strategy
cache_config = RequestCacheConfiguration()
cache_config.max_cached_versions = 3
# cache_config.recaching_strategy = ReCachingStrategy.cron(cron="* * * * *")
cache_config.recaching_strategy = ReCachingStrategy.when_requested()


req = c.create_request(url=remote_file_url, cache_configuration=cache_config)

# save requested file
req.download_response_content_to("/tmp/pubmed22n0003.xml.gz")

# or alternatively stream requested file
with open("/tmp/pubmed22n0003_2.xml.gz", "wb") as f:
    for chunk in req.download_response_content():
        f.write(chunk)



from concurrent.futures import thread
import requests
from email.utils import parsedate_to_datetime


def downloadtest():
    # url = "https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed22n1096.xml.gz"
    # url = "ftp://ftp.ncbi.nlm.nih.gov/pubmed/baseline/pubmed22n1096.xml.gz"
    # url = "http://ftp.ebi.ac.uk/pub/databases/GO/goa/HUMAN/goa_human.gaf.gz"
    url = "https://cloud.connect.dzd-ev.de/s/9fGr79Hwx6D7WZy/download"
    response = requests.head(url)
    print(response.status_code)
    print(response.headers.get("etag", None))
    # print(response.headers["ETag"])
    print(parsedate_to_datetime(response.headers["Date"]))
    print(requests.options(url).headers)
    """
    'Content-Length': '498430'
    'Date': 'Tue, 17 May 2022 12:39:51 GMT'
    'Content-Type': 'image/png'
    'Content-Disposition': 'attachment; filename*=UTF-8\'\'Dashboard.png; filename="Dashboard.png"'
    'Last-Modified': 'Sun, 12 Dec 2021 19:04:35 GMT'
    """


from pathlib import PurePath
from urllib.parse import urlparse


def filenameparser():
    url = "https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlparse"
    # url = "https://www.youtube.com/?v=QbTb9euNRWg"
    # url = "https://forum.duplicati.com/t/setting-up-amazon-s3/589"
    # url = "https://duplicacy.com/"
    print(PurePath(urlparse(url).path).suffix)


def pure_test():
    p = PurePath("/tmp/file.name")
    print(PurePath(str(p) + ".test"))


def catch_only_non_http_error():
    try:
        raise ValueError("MEH")
        raise requests.exceptions.TooManyRedirects("TOO MANY")
    except requests.exceptions.RequestException as e:
        print("THIS IS A HTTP ERROR")


def catch_thread_error():
    import threading
    import time
    import traceback

    class ThreadWithExpectionStackTraceStatus(threading.Thread):
        def __init__(self, *args, **kwargs):
            self.stack_trace: str = None
            self.failed: bool = False
            super().__init__(**kwargs)

        def run(self):
            try:
                self._target(*self._args, **self._kwargs)
            except:
                self.stack_trace = traceback.format_exc()
                self.failed = True
            finally:
                # Avoid a refcycle if the thread is running a function with
                # an argument that has a member that points to the thread.
                del self._target, self._args, self._kwargs

    def do_stuff():
        time.sleep(1)
        raise ValueError("Meehhooaaauuubooooom")

    t = ThreadWithExpectionStackTraceStatus(target=do_stuff)
    t.start()
    print(t.is_alive())
    print("STUFFF in MAIN")
    time.sleep(2)
    print("STUFFF2 in MAIN")
    try:
        print(t.is_alive())
        t.join()
        print(t.is_alive())
        if t.failed:
            print(t.stack_trace)
    except:
        print("Oh there was an error")


def print_error_msg():
    def make_error():
        raise ValueError("BLABLA")

    try:
        make_error()
    except ValueError as err:
        print(err)


def enumClassTest():
    from enum import Enum

    class TestEnum(str, Enum):
        class Strat:
            def __init__(self, param):
                self.param = param

        VAR = Strat(param=2)

    test = TestEnum.Strat.value

    print(type(test))
    print(test)
    print(test.param)

    # does not work!


def classnametest():
    import inspect

    class BaseWhat:
        def __eq__(self, other):
            if inspect.isclass(other):
                return other.__name__ == self.__class__.__name__
            return isinstance(other, self.__class__)

    class TestWhat(BaseWhat):
        pass

    class TestWhere(BaseWhat):
        pass

    instWhat = TestWhat()
    print(instWhat == TestWhat)
    instWhere = TestWhere()
    print(instWhat == instWhere)
    print(instWhere == TestWhat)
    print(instWhere == TestWhere)


def straStructTest():
    from pydantic import BaseModel
    from typing import Dict, Any, Union
    import inspect
    import json

    class ReCachingStrategy(BaseModel):
        strategy_name: str = None

        def __eq__(self, other):
            if inspect.isclass(other):
                return other.__name__ == self.__class__.__name__
            elif isinstance(other, str):
                return other == self.__class__.__name__
            return isinstance(other, self.__class__)

        def __init__(self, *args, **kwargs):
            super(ReCachingStrategy, self).__init__(*args, **kwargs)
            self.strategy_name = self.__class__.__name__

    class ReCachingStrategies:
        @classmethod
        def from_json(cls, strategy: Union[str, Dict]):
            if isinstance(strategy, str):
                strategy: Dict = json.loads(strategy)
            for strategy_class in cls.list():
                if strategy_class.__name__ == strategy["strategy_name"]:
                    return strategy_class(
                        **{k: v for k, v in strategy.items() if k != "strategy_name"}
                    )

        @classmethod
        def list(cls):
            return [
                cls_attribute
                for cls_attribute in cls.__dict__.values()
                if inspect.isclass(cls_attribute)
            ]

        class never(ReCachingStrategy):
            """never - Strategy
            Once downlaoded we will server only this cached file. no matter how often it is requested.
            This is usefull for files that will be 100% static.
            """

            pass

        class age(ReCachingStrategy):
            """age - Strategy
            Buffy will redownload when the cached version has a certain age in seconds. The default age is 3600sec
            This is usefull for dynamic remote resources, where the webserver wont provide any informations about the state of the resource like "etag", "last_modified_datetime" or "size_in_bytes"

            params:
                seconds - int - default: 3600 - The duration a resource will be cached until it gets a redownload from the source
            """

            seconds: int = 3600

    print(ReCachingStrategies.list())
    st_never = ReCachingStrategies.never()
    st_age = ReCachingStrategies.age()
    st_age_3 = ReCachingStrategies.age(seconds=3)
    print(st_age.json())
    print(st_age_3.json())
    print(st_never.json())
    st_age_3_b = ReCachingStrategies.from_json(st_age_3.json())
    print(st_age_3)
    print(st_age_3.json())


def straStructTest_v2():
    # https://pydantic-docs.helpmanual.io/usage/types/#discriminated-unions-aka-tagged-unions
    from pydantic import BaseModel, Field
    from typing import Dict, Any, Union, Literal
    import inspect

    class ReCachingStrategy(BaseModel):
        strategy_name: str = None

        def __eq__(self, other):
            if inspect.isclass(other):
                return other.__name__ == self.__class__.__name__
            elif isinstance(other, str):
                return other == self.__class__.__name__
            return isinstance(other, self.__class__)

        def __init__(self, *args, **kwargs):
            super(ReCachingStrategy, self).__init__(*args, **kwargs)
            self.strategy_name = self.__class__.__name__

    class ReCachingStrategies:
        @classmethod
        def list(cls):
            return [
                cls_attribute
                for cls_attribute in cls.__dict__.values()
                if inspect.isclass(cls_attribute)
            ]

        class never(ReCachingStrategy):
            """never - Strategy
            Once downlaoded we will server only this cached file. no matter how often it is requested.
            This is usefull for files that will be 100% static.
            """

            # strategy_name: Literal["never"] = "never"

            pass

        class age(ReCachingStrategy):
            """age - Strategy
            Buffy will redownload when the cached version has a certain age in seconds. The default age is 3600sec
            This is usefull for dynamic remote resources, where the webserver wont provide any informations about the state of the resource like "etag", "last_modified_datetime" or "size_in_bytes"

            params:
                seconds - int - default: 3600 - The duration a resource will be cached until it gets a redownload from the source
            """

            # strategy_name: Literal["age"] = "age"
            seconds: int = 3600

    class CachePlan(BaseModel):
        strategy: Union[ReCachingStrategies.never, ReCachingStrategies.age] = Field(
            discriminator="strategy_name"
        )

    print(ReCachingStrategies.list())
    st_never = ReCachingStrategies.never()
    st_age = ReCachingStrategies.age()
    st_age_3 = ReCachingStrategies.age(seconds=3)
    print("COMPARE TO CLASS POS", st_age == ReCachingStrategies.age)
    print("COMPARE TO CLASS NEG", st_age == ReCachingStrategies.never)
    print("CALL CLASS ATTR", ReCachingStrategies.age.seconds)
    print(st_age.json())
    print(st_age_3.json())
    print(st_never.json())
    # st_age_3_b = ReCachingStrategies.from_json(st_age_3.json())
    print(st_age_3)
    print(st_age_3.json())
    plan = CachePlan(strategy=st_age_3)
    print(plan.json())
    plan_parsed = CachePlan.parse_raw(plan.json())
    print(plan_parsed)


def pydantic_args_test():
    from pydantic import BaseModel

    class ReCachingStrategy(BaseModel):

        strategy_name: str = None

    r = ReCachingStrategy(2)
    print(r)
    print(r.json())


def pydantic_dyn_default_test():
    from pydantic import BaseModel, Field

    class naming_bootstrap(BaseModel):
        @classmethod
        def _get_strategy_name(cls) -> str:
            return cls.__name__

    class ReCachingStrategy(naming_bootstrap):

        strategy_name: str = Field(default_factory=naming_bootstrap._get_strategy_name)

        def __init__(self, *args, **kwargs):
            super(ReCachingStrategy, self).__init__(*args, **kwargs)
            # self.strategy_name = Field(self.__class__.__name__, const=True)
            # self.strategy_name = self.__class__.__name__

    r = ReCachingStrategy()
    print(r)  # nope -> strategy_name='naming_bootstrap'


def hashtest():
    import hashlib

    hashlib._Hash
    s = hashlib.sha256()
    print(type(s))


def file_seek_current_pos():
    f = open("README.md")
    f.seek(20)
    print(f.tell())
    f.close()


def is_file_like():
    from tempfile import TemporaryFile
    import io

    io.BufferedRandom

    t = TemporaryFile()
    print(t.tell())
    print(type(t))
    print(issubclass(type(t), io.BufferedIOBase))
    t.close()


def urlparse():
    from urllib.parse import urlparse
    from typing import Callable

    u = urlparse("google.de")
    for attr in dir(u):
        if not attr.startswith("_") and not isinstance(getattr(u, attr), Callable):
            print(attr, getattr(u, attr))


def catch_file_not_exists():
    file = "/tmp/nooooo"
    with open(
        file,
        "rb",
    ) as f:
        while True:
            chunk = f.read(1024)
            if chunk:
                yield chunk
            else:
                break


def redis_url_to_kwargs():
    import redis

    r = redis.Redis.from_url("redis://localhost:6379/0")
    print(r.get_connection_kwargs())


redis_url_to_kwargs()



from io import TextIOWrapper, FileIO
from typing import (
    Dict,
    List,
    overload,
    Generator,
    Literal,
    AsyncGenerator,
)
import redis
from pathlib import Path
import hashlib


from buffy.buffyserver.api.v1.models import (
    Response,
    Request,
    Request_in,
    RequestCacheConfiguration,
)
from buffy.tools.utils import url_to_path_dir_name

"""Backend brainstorming notes

* Store large bynaries files
* provide hash per file

"""


class StorageInterface:
    def __init__(self, file_storage_dir: Path, config=None):
        raise NotImplementedError

    def test_connection(self) -> bool:
        raise NotImplementedError

    def list_requests(
        self,
        group_name: str = None,  # Request.group_name
    ) -> List[Request]:
        raise NotImplementedError

    def get_request(self, request_id: str) -> Request:
        """_summary_

        Args:
            request_id (str): _description_

        Returns:
            Request: _description_
        """
        raise NotImplementedError

    def create_request(self, request: Request) -> Request:

        raise NotImplementedError

    def update_request(self, request: Request) -> Request:
        raise NotImplementedError

    def list_responses(self, request_id: str) -> List[Response]:
        raise NotImplementedError

    def create_response(self, response: Response) -> Response:
        raise NotImplementedError

    def update_response(self, response: Response) -> Response:
        raise NotImplementedError

    @overload
    def get_response(self, response_id: str) -> Response:
        raise NotImplementedError

    @overload
    def get_response(self, request_id: str = None, version: str = "latest") -> Response:
        """_summary_

        Args:
            request_id (str, optional): _description_. Defaults to None.
            version (str, optional): _description_. Defaults to "latest".

        Raises:
            NotImplementedError: _description_

        Returns:
            Response: _description_
        """
        raise NotImplementedError

    def update_response(self, response: Response):
        raise NotImplementedError

    def delete_response(self, response: Response):
        raise NotImplementedError

    ## ↓ CONTENT/FILE HANDLING ↓

    @overload
    async def read_response_content(
        self,
        response: Response,
        as_text: bool = Literal[False],
    ) -> AsyncGenerator[bytes, None]:
        raise NotImplementedError

    async def read_response_content(
        self,
        response: Response,
        as_text: bool = Literal[True],
    ) -> AsyncGenerator[str, None]:

        raise NotImplementedError

    def write_response_content(
        self,
        response: Response,
    ) -> FileIO:
        raise NotImplementedError

    def update_response_content_writing(
        self,
        response: Response,
    ):
        raise NotImplementedError

    def close_response_content_writing(self):
        raise NotImplementedError

    def get_response_content_hash(self, response: Response, cached_hash: bool = True):
        raise NotImplementedError
